# Copyright 2021-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import os
import shutil

import pendulum
from airflow.models import Connection
from airflow.utils.state import State

from oaebu_workflows.onix_telescope.onix_telescope import OnixTelescope, OnixRelease
from oaebu_workflows.oaebu_partners import partner_from_str
from oaebu_workflows.config import test_fixtures_folder
from observatory.platform.api import get_dataset_releases
from observatory.platform.bigquery import bq_sharded_table_id
from observatory.platform.gcs import gcs_blob_name_from_path
from observatory.platform.sftp import SftpFolders
from observatory.platform.observatory_config import Workflow
from observatory.platform.observatory_environment import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    SftpServer,
    find_free_port,
    load_and_parse_json,
)


class TestOnixTelescope(ObservatoryTestCase):
    """Tests for the ONIX telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(TestOnixTelescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.sftp_root = "/"
        self.date_regex = "\\d{8}"
        self.sftp_port = find_free_port()

        # Test file
        fixtures_folder = test_fixtures_folder(workflow_module="onix_telescope")
        self.onix_xml_path = os.path.join(fixtures_folder, "20210330_CURTINPRESS_ONIX.xml")
        self.onix_json_path = os.path.join(fixtures_folder, "20210330_CURTINPRESS_ONIX.json")

    def test_dag_structure(self):
        """Test that the ONIX DAG has the correct structure."""
        dag = OnixTelescope(
            dag_id="onix_telescope",
            cloud_workspace=self.fake_cloud_workspace,
            sftp_root=self.sftp_root,
            date_regex=self.date_regex,
        ).make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["list_release_info"],
                "list_release_info": ["move_files_to_in_progress"],
                "move_files_to_in_progress": ["download"],
                "download": ["upload_downloaded"],
                "upload_downloaded": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["move_files_to_finished"],
                "move_files_to_finished": ["add_new_dataset_releases"],
                "add_new_dataset_releases": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the Geonames DAG can be loaded from a DAG bag."""
        env = ObservatoryEnvironment(
            workflows=[
                Workflow(
                    dag_id="onix",
                    name="ONIX Telescope",
                    class_name="oaebu_workflows.onix_telescope.onix_telescope.OnixTelescope",
                    cloud_workspace=self.fake_cloud_workspace,
                    kwargs=dict(date_regex=self.date_regex),
                )
            ],
        )
        with env.create():
            self.assert_dag_load_from_config("onix")

        # Errors should be raised if kwargs dict not supplied
        env.workflows[0].kwargs = {}
        with env.create():
            with self.assertRaises(AssertionError) as cm:
                self.assert_dag_load_from_config("onix")
            msg = cm.exception.args[0]
            self.assertTrue("missing 1 required keyword-only argument" in msg)
            self.assertTrue("date_regex" in msg)

    def test_telescope(self):
        """Test the ONIX telescope end to end."""
        # Setup Observatory environment
        env = ObservatoryEnvironment(
            self.project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        sftp_server = SftpServer(host="localhost", port=self.sftp_port)
        dataset_id = env.add_dataset()

        # Create the Observatory environment and run tests

        with env.create(), sftp_server.create() as sftp_root:
            # Setup Telescope
            execution_date = pendulum.datetime(year=2021, month=3, day=31)
            partner = partner_from_str("onix", metadata_partner=True)
            partner.bq_dataset_id = dataset_id
            telescope = OnixTelescope(
                dag_id="onix_telescope_test",
                cloud_workspace=env.cloud_workspace,
                sftp_root="/",
                date_regex=self.date_regex,
                metadata_partner=partner,
            )
            dag = telescope.make_dag()

            # Release settings
            release_date = pendulum.datetime(year=2021, month=3, day=30)

            # Add SFTP connection
            conn = Connection(conn_id=telescope.sftp_service_conn_id, uri=f"ssh://:password@localhost:{self.sftp_port}")
            env.add_connection(conn)
            with env.create_dag_run(dag, execution_date):
                # Test that all dependencies are specified: no error should be thrown
                ti = env.run_task(telescope.check_dependencies.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Add ONIX file to SFTP server
                local_sftp_folders = SftpFolders(telescope.dag_id, telescope.sftp_service_conn_id, sftp_root)
                os.makedirs(local_sftp_folders.upload, exist_ok=True)
                onix_file_name = os.path.basename(self.onix_xml_path)
                onix_file_dst = os.path.join(local_sftp_folders.upload, onix_file_name)
                shutil.copy(self.onix_xml_path, onix_file_dst)

                # Get release info from SFTP server and check that the correct release info is returned via Xcom
                ti = env.run_task(telescope.list_release_info.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                expected_release_info = [{"release_date": release_date, "file_name": onix_file_name}]
                release_info = ti.xcom_pull(
                    key=OnixTelescope.RELEASE_INFO,
                    task_ids=telescope.list_release_info.__name__,
                    include_prior_dates=False,
                )
                for release in release_info:
                    rdate = release["release_date"]
                    release["release_date"] = pendulum.parse(rdate)
                self.assertEqual(expected_release_info, release_info)

                release = OnixRelease(
                    dag_id=telescope.dag_id,
                    run_id=env.dag_run.run_id,
                    snapshot_date=release_date,
                    onix_file_name=onix_file_name,
                )

                # Test move file to in progress
                ti = env.run_task(telescope.move_files_to_in_progress.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                in_progress_path = os.path.join(local_sftp_folders.in_progress, release.onix_file_name)
                self.assertFalse(os.path.isfile(onix_file_dst))
                self.assertTrue(os.path.isfile(in_progress_path))

                # Test download
                ti = env.run_task(telescope.download.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_file_integrity(release.download_path, "28f85c488ab01b0cff769d9da6b4be24", "md5")

                # Test upload downloaded
                ti = env.run_task(telescope.upload_downloaded.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_blob_integrity(
                    env.download_bucket, gcs_blob_name_from_path(release.download_path), release.download_path
                )

                # Test transform
                ti = env.run_task(telescope.transform.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_file_integrity(release.transform_path, "2164a300", "gzip_crc")

                # Test upload to cloud storage
                ti = env.run_task(telescope.upload_transformed.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_blob_integrity(
                    env.transform_bucket, gcs_blob_name_from_path(release.transform_path), release.transform_path
                )

                # Test load into BigQuery
                ti = env.run_task(telescope.bq_load.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                table_id = bq_sharded_table_id(
                    telescope.cloud_workspace.project_id,
                    telescope.metadata_partner.bq_dataset_id,
                    telescope.metadata_partner.bq_table_name,
                    release.snapshot_date,
                )
                self.assert_table_integrity(table_id, expected_rows=1)
                self.assert_table_content(table_id, load_and_parse_json(self.onix_json_path), primary_key="ISBN13")

                # Test move files to finished
                ti = env.run_task(telescope.move_files_to_finished.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                finished_path = os.path.join(local_sftp_folders.finished, onix_file_name)
                self.assertFalse(os.path.isfile(local_sftp_folders.in_progress))
                self.assertTrue(os.path.isfile(finished_path))

                # Add_dataset_release_task
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 0)
                ti = env.run_task(telescope.add_new_dataset_releases.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 1)

                # Test cleanup
                ti = env.run_task(telescope.cleanup.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_cleanup(release.workflow_folder)



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import gzip
import json
import logging
import os
import time
from typing import Dict, List, Optional, Tuple, Union

import pendulum
import requests
from airflow.exceptions import AirflowException, AirflowSkipException
from airflow.hooks.base import BaseHook
from google.auth import environment_vars
from google.auth.transport.requests import AuthorizedSession
from google.cloud.bigquery import TimePartitioningType, SourceFormat, WriteDisposition
from google.oauth2.service_account import IDTokenCredentials
from googleapiclient.discovery import Resource, build
from googleapiclient.errors import HttpError
from oauth2client.service_account import ServiceAccountCredentials

from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str
from observatory.api.client.model.dataset_release import DatasetRelease
from observatory.platform.api import make_observatory_api
from observatory.platform.airflow import AirflowConns
from observatory.platform.files import get_file_hash, save_jsonl_gz
from observatory.platform.files import add_partition_date
from observatory.platform.bigquery import bq_load_table, bq_table_id, bq_create_dataset
from observatory.platform.observatory_config import CloudWorkspace
from observatory.platform.workflows.workflow import (
    PartitionRelease,
    Workflow,
    cleanup,
    set_task_state,
    check_workflow_inputs,
)
from observatory.platform.gcs import (
    gcs_copy_blob,
    gcs_create_bucket,
    gcs_download_blob,
    gcs_upload_file,
    gcs_upload_files,
    gcs_blob_uri,
    gcs_blob_name_from_path,
)


class IrusOapenRelease(PartitionRelease):
    def __init__(
        self,
        dag_id: str,
        run_id: str,
        data_interval_start: pendulum.DateTime,
        data_interval_end: pendulum.DateTime,
        partition_date: pendulum.DateTime,
    ):
        """Create a IrusOapenRelease instance.

        :param dag_id: The ID of the DAG
        :param run_id: The Airflow run ID
        :param partition_date: The date of the partition/release
        """
        super().__init__(dag_id=dag_id, run_id=run_id, partition_date=partition_date)
        self.data_interval_start = data_interval_start
        self.data_interval_end = data_interval_end
        self.download_path = os.path.join(self.download_folder, "irus_oapen.jsonl.gz")
        self.transform_path = os.path.join(self.transform_folder, "irus_oapen.jsonl.gz")
        self.blob_name = gcs_blob_name_from_path(
            os.path.join(self.download_folder, f'{self.partition_date.format("YYYY_MM")}.jsonl.gz')
        )
        self.cloud_function_path = os.path.join(self.download_folder, "oapen_cloud_function.zip")


class IrusOapenTelescope(Workflow):
    OAPEN_PROJECT_ID = "oapen-usage-data-gdpr-proof"  # The oapen project id.
    OAPEN_BUCKET = f"{OAPEN_PROJECT_ID}_cloud-function"  # Storage bucket with the source code
    FUNCTION_NAME = "oapen-access-stats"  # Name of the google cloud function
    FUNCTION_REGION = "europe-west1"  # Region of the google cloud function
    FUNCTION_SOURCE_URL = (
        "https://github.com/The-Academic-Observatory/oapen-irus-uk-cloud-function/releases/"
        "download/v1.1.9/oapen-irus-uk-cloud-function.zip"
    )  # URL to the zipped source code of the cloud function
    FUNCTION_MD5_HASH = "946bb4d7ca229b15aba36ad7b5ed56d0"  # MD5 hash of the zipped source code
    FUNCTION_BLOB_NAME = "cloud_function_source_code.zip"  # blob name of zipped source code
    FUNCTION_TIMEOUT = 1500  # Timeout of cloud function in seconds. Maximum of 60 minutes,
    # see https://cloud.google.com/functions/docs/2nd-gen/overview#enhanced_infrastructure

    def __init__(
        self,
        dag_id: str,
        cloud_workspace: CloudWorkspace,
        publisher_name_v4: str,
        publisher_uuid_v5: str,
        data_partner: Union[str, OaebuPartner] = "irus_oapen",
        bq_dataset_description: str = "IRUS dataset",
        bq_table_description: str = None,
        api_dataset_id: str = "oapen",
        max_cloud_function_instances: int = 0,
        observatory_api_conn_id: str = AirflowConns.OBSERVATORY_API,
        geoip_license_conn_id: str = "geoip_license_key",
        irus_oapen_api_conn_id: str = "irus_api",
        irus_oapen_login_conn_id: str = "irus_login",
        catchup: bool = True,
        start_date: pendulum.DateTime = pendulum.datetime(2015, 6, 1),
        schedule: str = "0 0 4 * *",  # Run on the 4th of every month
        max_active_runs: int = 5,
    ):
        """The OAPEN irus uk telescope.
        :param dag_id: The ID of the DAG
        :param cloud_workspace: The CloudWorkspace object for this DAG
        :param publisher_name_v4: The publisher's name for version 4
        :param publisher_uuid_v5: The publisher's uuid for version 5
        :param data_partner: The data partner
        :param bq_dataset_description: Description for the BigQuery dataset
        :param bq_table_description: Description for the biguery table
        :param api_dataset_id: The ID to store the dataset release in the API
        :param max_cloud_function_instances:
        :param observatory_api_conn_id: Airflow connection ID for the overvatory API
        :param geoip_license_conn_id: The Airflow connection ID for the GEOIP license
        :param irus_oapen_api_conn_id: The Airflow connection ID for IRUS API - for counter 5
        :param irus_oapen_login_conn_id: The Airflow connection ID for IRUS API (login) - for counter 4
        :param catchup: Whether to catchup the DAG or not
        :param start_date: The start date of the DAG
        :param schedule: The schedule interval of the DAG
        :param max_active_runs: The maximum number of concurrent DAG instances
        """
        if bq_table_description is None:
            bq_table_description = "OAPEN metrics as recorded by the IRUS platform"

        super().__init__(
            dag_id,
            start_date,
            schedule,
            catchup=catchup,
            airflow_conns=[
                observatory_api_conn_id,
                geoip_license_conn_id,
                irus_oapen_api_conn_id,
                irus_oapen_login_conn_id,
            ],
            max_active_runs=max_active_runs,
            tags=["oaebu"],
        )
        self.dag_id = dag_id
        self.cloud_workspace = cloud_workspace
        self.publisher_name_v4 = publisher_name_v4
        self.publisher_uuid_v5 = publisher_uuid_v5
        self.data_partner = partner_from_str(data_partner)
        self.bq_dataset_description = bq_dataset_description
        self.bq_table_description = bq_table_description
        self.api_dataset_id = api_dataset_id
        self.max_cloud_function_instances = max_cloud_function_instances
        self.observatory_api_conn_id = observatory_api_conn_id
        self.geoip_license_conn_id = geoip_license_conn_id
        self.irus_oapen_api_conn_id = irus_oapen_api_conn_id
        self.irus_oapen_login_conn_id = irus_oapen_login_conn_id

        check_workflow_inputs(self)

        self.add_setup_task(self.check_dependencies)
        # create PythonOperator with task concurrency of 1, so tasks to create cloud function never run in parallel
        self.add_task(self.create_cloud_function, task_concurrency=1)
        self.add_task(self.call_cloud_function)
        self.add_task(self.transfer)
        self.add_task(self.download_transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.add_new_dataset_releases)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> List[IrusOapenRelease]:
        """Create a list of IrusOapenRelease instances for a given month.
        Say the dag is scheduled to run on 2022-04-07
        Interval_start will be 2022-03-01
        Interval_end will be 2022-04-01
        partition_date will be 2022-03-31

        :param kwargs: the context passed from the PythonOperator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for the keyword arguments that can be passed
        :return: list of IrusOapenRelease instances
        """
        # Get release_date
        data_interval_start = kwargs["data_interval_start"].start_of("month")
        data_interval_end = kwargs["data_interval_end"].start_of("month")
        partition_date = data_interval_start.end_of("month")

        logging.info(f"Release/partition date: {partition_date}")
        releases = [
            IrusOapenRelease(
                dag_id=self.dag_id,
                run_id=kwargs["run_id"],
                data_interval_start=data_interval_start,
                data_interval_end=data_interval_end,
                partition_date=partition_date,
            )
        ]
        return releases

    def transfer(self, releases: List[IrusOapenRelease], **kwargs):
        """Task to transfer the file for each release.

        :param releases: the list of IrusOapenRelease instances.
        """
        for release in releases:
            success = gcs_copy_blob(
                blob_name=release.blob_name,
                src_bucket=IrusOapenTelescope.OAPEN_BUCKET,
                dst_bucket=self.cloud_workspace.download_bucket,
            )
            set_task_state(success, kwargs["ti"].task_id, release=release)

    def download_transform(self, releases: List[IrusOapenRelease], **kwargs):
        """Task to download the access stats to a local file for each release."""
        for release in releases:
            success = gcs_download_blob(
                bucket_name=self.cloud_workspace.download_bucket,
                blob_name=release.blob_name,
                file_path=release.download_path,
            )
            set_task_state(success, kwargs["ti"].task_id, release=release)

            # Read gzipped data and create list of dicts
            with gzip.open(release.download_path, "r") as f:
                results = [json.loads(line) for line in f]

            # Add partition date
            results = add_partition_date(
                results, release.partition_date, TimePartitioningType.MONTH, partition_field="release_date"
            )

            # Write list into gzipped JSON Lines file
            save_jsonl_gz(release.transform_path, results)

    def create_cloud_function(self, releases: List[IrusOapenRelease], **kwargs):
        """Task to create the cloud function for each release."""
        for release in releases:
            # set up cloud function variables
            oapen_project_id = IrusOapenTelescope.OAPEN_PROJECT_ID
            source_bucket = IrusOapenTelescope.OAPEN_BUCKET
            function_name = IrusOapenTelescope.FUNCTION_NAME
            function_region = IrusOapenTelescope.FUNCTION_REGION
            function_source_url = IrusOapenTelescope.FUNCTION_SOURCE_URL
            function_blob_name = IrusOapenTelescope.FUNCTION_BLOB_NAME
            location = f"projects/{oapen_project_id}/locations/{function_region}"
            full_name = f"{location}/functions/{function_name}"

            # zip source code and upload to bucket
            success, upload = upload_source_code_to_bucket(
                source_url=function_source_url,
                project_id=oapen_project_id,
                bucket_name=source_bucket,
                blob_name=function_blob_name,
                cloud_function_path=release.cloud_function_path,
            )
            set_task_state(success, kwargs["ti"].task_id, release=release)

            # initialise cloud functions api
            creds = ServiceAccountCredentials.from_json_keyfile_name(os.environ.get(environment_vars.CREDENTIALS))
            service = build(
                "cloudfunctions", "v2beta", credentials=creds, cache_discovery=False, static_discovery=False
            )

            # update or create cloud function
            exists = cloud_function_exists(service, full_name)
            if not exists or upload is True:
                update = True if exists else False
                success, msg = create_cloud_function(
                    service,
                    location,
                    full_name,
                    source_bucket,
                    function_blob_name,
                    self.max_cloud_function_instances,
                    update,
                )
                set_task_state(success, kwargs["ti"].task_id, release=release)
                logging.info(f"Creating or patching cloud function successful, response: {msg}")
            else:
                logging.info(f"Using existing cloud function, source code has not changed.")

    def call_cloud_function(self, releases: List[IrusOapenRelease], **kwargs):
        """Task to call the cloud function for each release."""
        for release in releases:
            # set up cloud function variables
            oapen_project_id = IrusOapenTelescope.OAPEN_PROJECT_ID
            source_bucket = IrusOapenTelescope.OAPEN_BUCKET
            function_name = IrusOapenTelescope.FUNCTION_NAME
            function_region = IrusOapenTelescope.FUNCTION_REGION
            location = f"projects/{oapen_project_id}/locations/{function_region}"
            full_name = f"{location}/functions/{function_name}"
            geoip_license_key = BaseHook.get_connection(self.geoip_license_conn_id).password

            # get the publisher_uuid or publisher_id, both are set to empty strings when publisher id is 'oapen'
            if release.partition_date >= pendulum.datetime(2020, 4, 1):
                airflow_conn = self.irus_oapen_api_conn_id
            else:
                airflow_conn = self.irus_oapen_login_conn_id
            username = BaseHook.get_connection(airflow_conn).login
            password = BaseHook.get_connection(airflow_conn).password

            # initialise cloud functions api
            creds = ServiceAccountCredentials.from_json_keyfile_name(os.environ.get(environment_vars.CREDENTIALS))
            service = build(
                "cloudfunctions", "v2beta", credentials=creds, cache_discovery=False, static_discovery=False
            )

            # Get cloud function uri
            function_uri = cloud_function_exists(service, full_name)

            call_cloud_function(
                function_uri,
                release.partition_date.format("YYYY-MM"),
                username,
                password,
                geoip_license_key,
                self.publisher_name_v4,
                self.publisher_uuid_v5,
                source_bucket,
                release.blob_name,
            )

    def upload_transformed(self, releases: List[IrusOapenRelease], **kwargs) -> None:
        """Uploads the transformed files to GCS for each release"""
        for release in releases:
            success = gcs_upload_files(
                bucket_name=self.cloud_workspace.transform_bucket,
                file_paths=[release.transform_path],
            )
            set_task_state(success, kwargs["ti"].task_id, release=release)

    def bq_load(self, releases: List[IrusOapenRelease], **kwargs) -> None:
        """Loads the sales and traffic data into BigQuery"""
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.data_partner.bq_dataset_id,
            location=self.cloud_workspace.data_location,
            description=self.bq_dataset_description,
        )
        for release in releases:
            uri = gcs_blob_uri(self.cloud_workspace.transform_bucket, gcs_blob_name_from_path(release.transform_path))
            table_id = bq_table_id(
                self.cloud_workspace.project_id, self.data_partner.bq_dataset_id, self.data_partner.bq_table_name
            )
            state = bq_load_table(
                uri=uri,
                table_id=table_id,
                schema_file_path=self.data_partner.schema_path,
                source_format=SourceFormat.NEWLINE_DELIMITED_JSON,
                partition_type=TimePartitioningType.MONTH,
                partition=True,
                partition_field="release_date",
                write_disposition=WriteDisposition.WRITE_APPEND,
                table_description=self.bq_table_description,
                ignore_unknown_values=True,
            )
            set_task_state(state, kwargs["ti"].task_id, release=release)

    def add_new_dataset_releases(self, releases: List[IrusOapenRelease], **kwargs) -> None:
        """Adds release information to API."""
        api = make_observatory_api(observatory_api_conn_id=self.observatory_api_conn_id)
        for release in releases:
            dataset_release = DatasetRelease(
                dag_id=self.dag_id,
                dataset_id=self.api_dataset_id,
                dag_run_id=release.run_id,
                data_interval_start=release.data_interval_start,
                data_interval_end=release.data_interval_end,
                partition_date=release.partition_date,
            )
            api.post_dataset_release(dataset_release)

    def cleanup(self, releases: List[IrusOapenRelease], **kwargs) -> None:
        """Delete all files, folders and XComs associated with this release."""
        for release in releases:
            cleanup(
                dag_id=self.dag_id, execution_date=kwargs["execution_date"], workflow_folder=release.workflow_folder
            )


def upload_source_code_to_bucket(
    source_url: str, project_id: str, bucket_name: str, blob_name: str, cloud_function_path: str
) -> Tuple[bool, bool]:
    """Upload source code of cloud function to storage bucket

    :param source_url: The url to the zip file with source code
    :param project_id: The project id with the bucket
    :param bucket_name: The bucket name
    :param blob_name: The blob name
    :param cloud_function_path: The local path to the cloud function
    :return: Whether task was successful and whether file was uploaded
    """

    # Get zip file with source code from github release
    response = requests.get(source_url)
    with open(cloud_function_path, "wb") as f:
        f.write(response.content)

    # Check if current md5 hash matches expected md5 hash
    expected_md5_hash = IrusOapenTelescope.FUNCTION_MD5_HASH
    actual_md5_hash = get_file_hash(file_path=cloud_function_path, algorithm="md5")
    if expected_md5_hash != actual_md5_hash:
        raise AirflowException(f"md5 hashes do not match, expected: {expected_md5_hash}, actual: {actual_md5_hash}")

    # Create storage bucket
    gcs_create_bucket(bucket_name=bucket_name, location="EU", project_id=project_id, lifecycle_delete_age=1)

    # upload zip to cloud storage
    success, upload = gcs_upload_file(
        bucket_name=bucket_name, blob_name=blob_name, file_path=cloud_function_path, project_id=project_id
    )
    return success, upload


def cloud_function_exists(service: Resource, full_name: str) -> Optional[str]:
    """Check if cloud function with a given name already exists

    :param service: Cloud function service
    :param full_name: Name of the cloud function
    :return: URI if cloud function exists, else None
    """
    try:
        response = service.projects().locations().functions().get(name=full_name).execute()
        uri = response["serviceConfig"]["uri"]
    except HttpError:
        return None
    return uri


def create_cloud_function(
    service: Resource,
    location: str,
    full_name: str,
    source_bucket: str,
    blob_name: str,
    max_active_runs: int,
    update: bool,
) -> Tuple[bool, dict]:
    """Create cloud function.

    :param service: Cloud function service
    :param location: Location of the cloud function
    :param full_name: Name of the cloud function
    :param source_bucket: Name of bucket where the source code is stored
    :param blob_name: Blob name of source code inside bucket
    :param max_active_runs: The limit on the maximum number of function instances that may coexist at a given time
    :param update: Whether a new function is created or an existing one is updated
    :return: Status of the cloud function and error/success message
    """
    body = {
        "name": full_name,
        "environment": "GEN_2",
        "description": "Pulls oapen irus uk data and replaces ip addresses with city and country info.",
        "buildConfig": {
            "runtime": "python39",
            "entryPoint": "download",
            "source": {"storageSource": {"bucket": source_bucket, "object": blob_name}},
        },
        "serviceConfig": {
            "timeoutSeconds": IrusOapenTelescope.FUNCTION_TIMEOUT,
            "availableMemory": "4096M",
            "maxInstanceCount": max_active_runs,
            "allTrafficOnLatestRevision": True,
        },
    }
    if update:
        update_mask = ",".join(body.keys())
        response = (
            service.projects()
            .locations()
            .functions()
            .patch(name=full_name, updateMask=update_mask, body=body)
            .execute()
        )
        logging.info(f"Patching cloud function, response: {response}")
    else:
        response = (
            service.projects()
            .locations()
            .functions()
            .create(parent=location, functionId=IrusOapenTelescope.FUNCTION_NAME, body=body)
            .execute()
        )
        logging.info(f"Creating cloud function, response: {response}")

    operation_name = response.get("name")
    done = response.get("done")
    while not done:
        time.sleep(10)
        response = service.projects().locations().operations().get(name=operation_name).execute()
        done = response.get("done")

    error = response.get("error")
    response = response.get("response")
    if response:
        msg = response
        success = True
    else:
        msg = error
        success = False

    return success, msg


def call_cloud_function(
    function_uri: str,
    release_date: str,
    username: str,
    password: str,
    geoip_license_key: str,
    publisher_name_v4: str,
    publisher_uuid_v5: str,
    bucket_name: str,
    blob_name: str,
) -> None:
    """Iteratively call cloud function, until it has finished processing all publishers.
    When a publisher name/uuid  is given, there is only 1 publisher, if it is empty the cloud function will process
    all available publishers. In that case, when the data is downloaded from the new platform it can be done in 1
    iteration, however for the old platform two files have to be downloaded separately for each publisher,
    this might take longer than the timeout time of the cloud function, so the process is split up in multiple calls.

    :param function_uri: URI of the cloud function
    :param release_date: The release date in YYYY-MM
    :param username: Oapen username (email or requestor_id)
    :param password: Oapen password (password or api_key)
    :param geoip_license_key: License key of geoip database
    :param publisher_name_v4: URL encoded name of the publisher (used for counter version 4)
    :param publisher_uuid_v5: UUID of the publisher (used for counter version 5)
    :param bucket_name: Name of the bucket to store oapen access stats data
    :param blob_name: Blob name to store oapen access stats data
    """
    creds = IDTokenCredentials.from_service_account_file(
        os.environ.get(environment_vars.CREDENTIALS), target_audience=function_uri
    )
    authed_session = AuthorizedSession(creds)
    data = {
        "release_date": release_date,
        "username": username,
        "password": password,
        "geoip_license_key": geoip_license_key,
        "publisher_name_v4": publisher_name_v4,
        "publisher_uuid_v5": publisher_uuid_v5,
        "bucket_name": bucket_name,
        "blob_name": blob_name,
    }
    finished = False
    while not finished:
        response = authed_session.post(
            function_uri,
            data=json.dumps(data),
            headers={"Content-Type": "application/json"},
            timeout=IrusOapenTelescope.FUNCTION_TIMEOUT,
        )
        logging.info(f"Call cloud function response status code: {response.status_code}, reason: {response.reason}")
        if response.status_code != 200:
            raise AirflowException("Cloud function unsuccessful")

        response_json = response.json()
        if response_json["unprocessed_publishers"]:
            data["unprocessed_publishers"] = response_json["unprocessed_publishers"]
            remaining_publishers = len(response_json["unprocessed_publishers"])
        else:
            finished = True
            remaining_publishers = 0

        entries = response_json["entries"]
        if entries == 0 and remaining_publishers == 0:
            raise AirflowSkipException("No access stats entries for publisher(s) in month.")

        logging.info(f"Processed {entries} entries in total. {remaining_publishers} publishers " f"left to process")



# Copyright 2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Keegan Smith

import os
import logging
from typing import Union

import pendulum
from airflow.exceptions import AirflowException
from google.cloud.bigquery import SourceFormat

from oaebu_workflows.onix_utils import (
    onix_parser_download,
    onix_parser_execute,
    collapse_subjects,
    create_personname_fields,
)
from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str
from observatory.api.client.model.dataset_release import DatasetRelease
from observatory.platform.api import make_observatory_api
from observatory.platform.airflow import AirflowConns
from observatory.platform.bigquery import bq_load_table, bq_sharded_table_id, bq_create_dataset
from observatory.platform.observatory_config import CloudWorkspace
from observatory.platform.utils.url_utils import retry_get_url
from observatory.platform.files import save_jsonl_gz, load_jsonl
from observatory.platform.gcs import gcs_upload_files, gcs_blob_name_from_path, gcs_blob_uri
from observatory.platform.workflows.workflow import (
    Workflow,
    SnapshotRelease,
    make_snapshot_date,
    cleanup,
    set_task_state,
    check_workflow_inputs,
)


THOTH_URL = "{host_name}/specifications/{format_specification}/publisher/{publisher_id}"
DEFAULT_HOST_NAME = "https://export.thoth.pub"


class ThothRelease(SnapshotRelease):
    def __init__(
        self,
        *,
        dag_id: str,
        run_id: str,
        snapshot_date: pendulum.DateTime,
    ):
        """Construct a ThothRelease.
        :param dag_id: The ID of the DAG
        :param run_id: The Airflow run ID
        :param release_date: The date of the snapshot_date/release
        """
        super().__init__(dag_id=dag_id, run_id=run_id, snapshot_date=snapshot_date)
        self.download_path = os.path.join(self.download_folder, f"thoth_{snapshot_date.format('YYYY_MM_DD')}.xml")
        self.transform_path = os.path.join(self.download_folder, f"thoth_{snapshot_date.format('YYYY_MM_DD')}.jsonl.gz")


class ThothTelescope(Workflow):
    def __init__(
        self,
        *,
        dag_id: str,
        cloud_workspace: CloudWorkspace,
        publisher_id: str,
        format_specification: str,
        metadata_partner: Union[str, OaebuPartner] = "thoth",
        bq_dataset_description: str = "Thoth ONIX Feed",
        bq_table_description: str = None,
        api_dataset_id: str = "onix",
        host_name: str = "https://export.thoth.pub",
        observatory_api_conn_id: str = AirflowConns.OBSERVATORY_API,
        catchup: bool = False,
        start_date: pendulum.DateTime = pendulum.datetime(2022, 12, 1),
        schedule: str = "@weekly",
    ):
        """Construct an ThothOnixTelescope instance.
        :param dag_id: The ID of the DAG
        :param cloud_workspace: The CloudWorkspace object for this DAG
        :param publisher_id: The Thoth ID for this publisher
        :param format_specification: The Thoth ONIX/metadata format specification. e.g. "onix_3.0::oapen"
        :param metadata_partner: The metadata partner name
        :param bq_dataset_description: Description for the BigQuery dataset
        :param bq_table_description: Description for the biguery table
        :param api_dataset_id: The ID to store the dataset release in the API
        :param host_name: The Thoth host name
        :param observatory_api_conn_id: Airflow connection ID for the overvatory API
        :param catchup: Whether to catchup the DAG or not
        :param start_date: The start date of the DAG
        :param schedule: The schedule interval of the DAG
        """
        super().__init__(
            dag_id,
            start_date=start_date,
            schedule=schedule,
            airflow_conns=[observatory_api_conn_id],
            catchup=catchup,
            tags=["oaebu"],
        )

        self.dag_id = dag_id
        self.cloud_workspace = cloud_workspace
        self.publisher_id = publisher_id
        self.metadata_partner = partner_from_str(metadata_partner, metadata_partner=True)
        self.bq_dataset_description = bq_dataset_description
        self.bq_table_description = bq_table_description
        self.api_dataset_id = api_dataset_id
        self.host_name = host_name
        self.format_specification = format_specification
        self.observatory_api_conn_id = observatory_api_conn_id

        check_workflow_inputs(self)

        self.add_setup_task(self.check_dependencies)
        self.add_task(self.download)
        self.add_task(self.upload_downloaded)
        self.add_task(self.transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.add_new_dataset_releases)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> ThothRelease:
        """Creates a new Thoth release instance

        :param kwargs: the context passed from the PythonOperator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for the keyword arguments that can be passed
        :return: The Thoth release instance
        """
        snapshot_date = make_snapshot_date(**kwargs)
        release = ThothRelease(dag_id=self.dag_id, run_id=kwargs["run_id"], snapshot_date=snapshot_date)
        return release

    def download(self, release: ThothRelease, **kwargs) -> None:
        """Task to download the ONIX release from Thoth.

        :param release: The Thoth release instance
        """
        thoth_download_onix(
            publisher_id=self.publisher_id,
            format_spec=self.format_specification,
            download_path=release.download_path,
        )

    def upload_downloaded(self, release: ThothRelease, **kwargs) -> None:
        """Upload the downloaded thoth onix XML to google cloud bucket"""
        success = gcs_upload_files(bucket_name=self.cloud_workspace.download_bucket, file_paths=[release.download_path])
        set_task_state(success, kwargs["ti"].task_id, release=release)

    def transform(self, release: ThothRelease, **kwargs) -> None:
        """Task to transform the Thoth ONIX data"""
        success, parser_path = onix_parser_download()
        set_task_state(success, task_id=kwargs["ti"].task_id, release=release)
        success = onix_parser_execute(
            parser_path, input_dir=release.download_folder, output_dir=release.transform_folder
        )
        set_task_state(success, task_id=kwargs["ti"].task_id, release=release)
        transformed = collapse_subjects(load_jsonl(os.path.join(release.transform_folder, "full.jsonl")))
        transformed = create_personname_fields(transformed)
        save_jsonl_gz(release.transform_path, transformed)

    def upload_transformed(self, release: ThothRelease, **kwargs) -> None:
        """Upload the downloaded thoth onix .jsonl to google cloud bucket"""
        success = gcs_upload_files(
            bucket_name=self.cloud_workspace.transform_bucket, file_paths=[release.transform_path]
        )
        set_task_state(success, kwargs["ti"].task_id, release=release)

    def bq_load(self, release: ThothRelease, **kwargs) -> None:
        """Task to load the transformed ONIX jsonl file to BigQuery."""
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.metadata_partner.bq_dataset_id,
            location=self.cloud_workspace.data_location,
            description=self.bq_dataset_description,
        )
        uri = gcs_blob_uri(self.cloud_workspace.transform_bucket, gcs_blob_name_from_path(release.transform_path))
        table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.metadata_partner.bq_dataset_id,
            self.metadata_partner.bq_table_name,
            release.snapshot_date,
        )
        state = bq_load_table(
            uri=uri,
            table_id=table_id,
            schema_file_path=self.metadata_partner.schema_path,
            source_format=SourceFormat.NEWLINE_DELIMITED_JSON,
            table_description=self.bq_table_description,
        )
        set_task_state(state, kwargs["ti"].task_id, release=release)

    def add_new_dataset_releases(self, release: ThothRelease, **kwargs) -> None:
        """Adds release information to API."""
        dataset_release = DatasetRelease(
            dag_id=self.dag_id,
            dataset_id=self.api_dataset_id,
            dag_run_id=release.run_id,
            snapshot_date=release.snapshot_date,
            data_interval_start=kwargs["data_interval_start"],
            data_interval_end=kwargs["data_interval_end"],
        )
        api = make_observatory_api(observatory_api_conn_id=self.observatory_api_conn_id)
        api.post_dataset_release(dataset_release)

    def cleanup(self, release: ThothRelease, **kwargs) -> None:
        """Delete all files, folders and XComs associated with this release."""
        cleanup(dag_id=self.dag_id, execution_date=kwargs["execution_date"], workflow_folder=release.workflow_folder)


def thoth_download_onix(
    publisher_id: str,
    download_path: str,
    format_spec: str,
    host_name: str = DEFAULT_HOST_NAME,
    num_retries: int = 3,
) -> None:
    """Hits the Thoth API and requests the ONIX feed for a particular publisher.
    Creates a file called onix.xml at the specified location

    :param publisher_id: The ID of the publisher. Can be found using Thoth GraphiQL API
    :param download_path: The path to download ONIX the file to
    :param format_spec: The ONIX format specification to use. Options can be found with the /formats endpoint of the API
    :param host_name: The Thoth host URL
    :param num_retries: The number of times to retry the download, given an unsuccessful return code
    """
    url = THOTH_URL.format(host_name=host_name, format_specification=format_spec, publisher_id=publisher_id)
    logging.info(f"Downloading ONIX XML from {url}")
    response = retry_get_url(url, num_retries=num_retries)
    if response.status_code != 200:
        raise AirflowException(
            f"Request for URL {url} was unsuccessful with code: {response.status_code}\nContent response: {response.content.decode('utf-8')}"
        )
    with open(download_path, "wb") as f:
        f.write(response.content)



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, Keegan Smith

import gzip
import json
import os
from unittest.mock import patch

import pendulum
from airflow.models import Connection
from airflow.utils.state import State
from googleapiclient.discovery import build
from googleapiclient.http import HttpMockSequence

from oaebu_workflows.google_analytics3_telescope.google_analytics3_telescope import GoogleAnalytics3Telescope
from oaebu_workflows.oaebu_partners import partner_from_str
from oaebu_workflows.config import test_fixtures_folder
from observatory.platform.api import get_dataset_releases
from observatory.platform.observatory_config import Workflow
from observatory.platform.bigquery import bq_table_id
from observatory.platform.observatory_environment import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    find_free_port,
    load_and_parse_json,
)


class TestGoogleAnalytics3Telescope(ObservatoryTestCase):
    """Tests for the Google Analytics telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.
        :param args: arguments.
        :param kwargs: keyword arguments.
        """
        super(TestGoogleAnalytics3Telescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.view_id = "11235141"
        self.pagepath_regex = r".*regex$"
        self.organisation_name = "UCL Press"

        fixtures_folder = test_fixtures_folder(workflow_module="google_analytics3_telescope")
        self.test_table = os.path.join(fixtures_folder, "test_table.json")
        self.test_table_anu = os.path.join(fixtures_folder, "test_table_anu.json")

    def test_dag_structure(self):
        """Test that the Google Analytics DAG has the correct structure.
        :return: None
        """
        cloud_workspace = self.fake_cloud_workspace
        dag = GoogleAnalytics3Telescope(
            dag_id="google_analytics_test",
            organisation_name="Organisation Name",
            cloud_workspace=cloud_workspace,
            view_id=self.view_id,
            pagepath_regex=self.pagepath_regex,
        ).make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["download_transform"],
                "download_transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["add_new_dataset_releases"],
                "add_new_dataset_releases": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the Google Analytics DAG can be loaded from a DAG bag."""

        env = ObservatoryEnvironment(
            workflows=[
                Workflow(
                    dag_id="google_analytics3",
                    name="My Google Analytics Workflow",
                    class_name="oaebu_workflows.google_analytics3_telescope.google_analytics3_telescope.GoogleAnalytics3Telescope",
                    cloud_workspace=self.fake_cloud_workspace,
                    kwargs=dict(organisation_name="My Organisation", pagepath_regex="", view_id="123456"),
                )
            ]
        )
        with env.create():
            self.assert_dag_load_from_config("google_analytics3")

        # Errors should be raised if kwargs dict not supplied
        env.workflows[0].kwargs = {}
        with env.create():
            with self.assertRaises(AssertionError) as cm:
                self.assert_dag_load_from_config("google_analytics3")
            msg = cm.exception.args[0]
            self.assertTrue("missing 3 required positional arguments" in msg)
            self.assertTrue("organisation_name" in msg)
            self.assertTrue("pagepath_regex" in msg)
            self.assertTrue("view_id" in msg)

    @patch("oaebu_workflows.google_analytics3_telescope.google_analytics3_telescope.build")
    @patch(
        "oaebu_workflows.google_analytics3_telescope.google_analytics3_telescope.ServiceAccountCredentials"
    )
    def test_telescope(self, mock_account_credentials, mock_build):
        """Test the Google Analytics telescope end to end specifically for ANU Press, to test custom dimensions.
        :return: None.
        """
        # Mock the Google Reporting Analytics API service
        mock_account_credentials.from_json_keyfile_dict.return_value = ""

        http = HttpMockSequence(create_http_mock_sequence(GoogleAnalytics3Telescope.ANU_ORG_NAME))
        mock_build.return_value = build("analyticsreporting", "v4", http=http)

        # Setup Observatory environment
        env = ObservatoryEnvironment(
            self.project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        # Setup Telescope
        execution_date = pendulum.datetime(year=2022, month=6, day=1)
        partner = partner_from_str("google_analytics3")
        partner.bq_dataset_id = env.add_dataset()
        telescope = GoogleAnalytics3Telescope(
            dag_id="google_analytics_test",
            organisation_name=GoogleAnalytics3Telescope.ANU_ORG_NAME,
            cloud_workspace=env.cloud_workspace,
            view_id=self.view_id,
            pagepath_regex=self.pagepath_regex,
            data_partner=partner,
        )
        dag = telescope.make_dag()

        # Create the Observatory environment and run tests
        with env.create():
            with env.create_dag_run(dag, execution_date):
                # Add OAEBU service account connection connection
                conn = Connection(
                    conn_id="oaebu_service_account",
                    uri=f"google-cloud-platform://?type=service_account&private_key_id=private_key_id"
                    f"&private_key=private_key"
                    f"&client_email=client_email"
                    f"&client_id=client_id",
                )
                env.add_connection(conn)

                # Test that all dependencies are specified: no error should be thrown
                ti = env.run_task(telescope.check_dependencies.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test download_transform task
                ti = env.run_task(telescope.download_transform.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test that transformed file uploaded
                ti = env.run_task(telescope.upload_transformed.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test that data loaded into BigQuery
                ti = env.run_task(telescope.bq_load.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Use release to check tasks
                release = telescope.make_release(
                    run_id=env.dag_run.run_id,
                    data_interval_start=pendulum.parse(str(env.dag_run.data_interval_start)),
                    data_interval_end=pendulum.parse(str(env.dag_run.data_interval_end)),
                )[0]

                # Test download_transform task
                self.assertTrue(os.path.exists(release.transform_path))
                self.assertTrue(os.path.isfile(release.transform_path))
                # Use frozenset to test results are as expected, many dict transformations re-order items in dict
                actual_list = []
                with gzip.open(release.transform_path, "rb") as f:
                    for line in f:
                        actual_list.append(json.loads(line))
                expected_list = [
                    {
                        "url": "/base/path/151420",
                        "title": "Anything public program drive north.",
                        "start_date": "2022-06-01",
                        "end_date": "2022-06-30",
                        "average_time": 59.5,
                        "unique_views": {
                            "country": [{"name": "country 1", "value": 3}, {"name": "country 2", "value": 3}],
                            "referrer": [{"name": "referrer 1", "value": 3}, {"name": "referrer 2", "value": 3}],
                            "social_network": [
                                {"name": "social_network 1", "value": 3},
                                {"name": "social_network 2", "value": 3},
                            ],
                        },
                        "page_views": {
                            "country": [{"name": "country 1", "value": 4}, {"name": "country 2", "value": 4}],
                            "referrer": [{"name": "referrer 1", "value": 4}, {"name": "referrer 2", "value": 4}],
                            "social_network": [
                                {"name": "social_network 1", "value": 4},
                                {"name": "social_network 2", "value": 4},
                            ],
                        },
                        "sessions": {
                            "country": [{"name": "country 1", "value": 1}, {"name": "country 2", "value": 1}],
                            "source": [{"name": "source 1", "value": 1}, {"name": "source 2", "value": 1}],
                        },
                        "publication_id": "1234567890123",
                        "publication_type": "book",
                        "publication_imprint": "imprint",
                        "publication_group": "group",
                        "publication_whole_or_part": "whole",
                        "publication_format": "PDF",
                        "release_date": "2022-06-30",
                    },
                    {
                        "url": "/base/path/833557",
                        "title": "Standard current never no.",
                        "start_date": "2022-06-01",
                        "end_date": "2022-06-30",
                        "average_time": 49.6,
                        "unique_views": {"country": [], "referrer": [], "social_network": []},
                        "page_views": {"country": [], "referrer": [], "social_network": []},
                        "sessions": {"country": [], "source": []},
                        "publication_id": "1234567891234",
                        "publication_type": "book",
                        "publication_imprint": "imprint",
                        "publication_group": "(none)",
                        "publication_whole_or_part": "part",
                        "publication_format": "HTML",
                        "release_date": "2022-06-30",
                    },
                    {
                        "url": "/base/path/833557?fbclid=123",
                        "title": "Standard current never no.",
                        "start_date": "2022-06-01",
                        "end_date": "2022-06-30",
                        "average_time": 38.8,
                        "unique_views": {
                            "country": [{"name": "country 2", "value": 2}],
                            "referrer": [{"name": "referrer 2", "value": 2}],
                            "social_network": [{"name": "social_network 2", "value": 2}],
                        },
                        "page_views": {
                            "country": [{"name": "country 2", "value": 4}],
                            "referrer": [{"name": "referrer 2", "value": 4}],
                            "social_network": [{"name": "social_network 2", "value": 4}],
                        },
                        "sessions": {"country": [], "source": []},
                        "publication_id": "1234567891234",
                        "publication_type": "book",
                        "publication_imprint": "imprint",
                        "publication_group": "(none)",
                        "publication_whole_or_part": "part",
                        "publication_format": "HTML",
                        "release_date": "2022-06-30",
                    },
                ]
                self.assertEqual(3, len(actual_list))
                self.assertEqual(frozenset(expected_list[0]), frozenset(actual_list[0]))
                self.assertEqual(frozenset(expected_list[1]), frozenset(actual_list[1]))
                self.assertEqual(frozenset(expected_list[2]), frozenset(actual_list[2]))

                # Test that data loaded into BigQuery
                table_id = bq_table_id(
                    telescope.cloud_workspace.project_id,
                    telescope.data_partner.bq_dataset_id,
                    telescope.data_partner.bq_table_name,
                )
                self.assert_table_integrity(table_id, expected_rows=3)
                self.assert_table_content(
                    table_id,
                    load_and_parse_json(self.test_table_anu, date_fields=["release_date", "start_date", "end_date"]),
                    primary_key="url",
                )

                # add_dataset_release_task
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 0)
                ti = env.run_task(telescope.add_new_dataset_releases.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 1)

                # Test that all telescope data deleted
                ti = env.run_task(telescope.cleanup.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_cleanup(release.workflow_folder)


def create_http_mock_sequence(organisation_name: str) -> list:
    """Create a list of http mock sequences for listing books and getting dimension data

    :param organisation_name: The organisation name (add custom dimensions for ANU)
    :return: A list with HttpMockSequence instances
    """
    http_mock_sequence = []
    list_books = {
        "reports": [
            {
                "columnHeader": {
                    "dimensions": ["ga:pagepath", "ga:pageTitle"],
                    "metricHeader": {"metricHeaderEntries": [{"name": "ga:avgTimeOnPage", "type": "TIME"}]},
                },
                "data": {
                    "rows": [
                        {
                            "dimensions": ["/base/path/151420", "Anything public program drive north."],
                            "metrics": [{"values": ["59.5"]}],
                        },
                        {
                            "dimensions": ["/base/path/833557", "Standard current never no."],
                            "metrics": [{"values": ["49.6"]}],
                        },
                    ],
                    "totals": [{"values": ["109.1"]}],
                    "rowCount": 2,
                    "minimums": [{"values": ["49.6"]}],
                    "maximums": [{"values": ["59.5"]}],
                    "isDataGolden": True,
                },
                "nextPageToken": "200",
            }
        ]
    }
    # Add custom dimensions from ANU Press
    if organisation_name == GoogleAnalytics3Telescope.ANU_ORG_NAME:
        list_books["reports"][0]["columnHeader"]["dimensions"] += [f"ga:dimension{(str(i))}" for i in range(1, 7)]
        list_books["reports"][0]["data"]["rows"][0]["dimensions"] += [
            "1234567890123",
            "book",
            "imprint",
            "group",
            "whole",
            "PDF",
        ]
        list_books["reports"][0]["data"]["rows"][1]["dimensions"] += [
            "1234567891234",
            "book",
            "imprint",
            "(none)",
            "part",
            "HTML",
        ]
    list_books_next_page = {
        "reports": [
            {
                "columnHeader": {
                    "dimensions": ["ga:pagepath", "ga:pageTitle"],
                    "metricHeader": {"metricHeaderEntries": [{"name": "ga:avgTimeOnPage", "type": "TIME"}]},
                },
                "data": {
                    "rows": [
                        {
                            "dimensions": ["/base/path/833557?fbclid=123", "Standard current never no."],
                            "metrics": [{"values": ["38.8"]}],
                        }
                    ],
                    "totals": [{"values": ["38.8"]}],
                    "rowCount": 1,
                    "minimums": [{"values": ["38.8"]}],
                    "maximums": [{"values": ["38.8"]}],
                    "isDataGolden": True,
                },
            }
        ]
    }
    # Add custom dimensions from ANU Press
    if organisation_name == GoogleAnalytics3Telescope.ANU_ORG_NAME:
        list_books_next_page["reports"][0]["columnHeader"]["dimensions"] += [
            f"ga:dimension{(str(i))}" for i in range(1, 7)
        ]
        list_books_next_page["reports"][0]["data"]["rows"][0]["dimensions"] += [
            "1234567891234",
            "book",
            "imprint",
            "(none)",
            "part",
            "HTML",
        ]
    http_mock_sequence.append(({"status": "200"}, json.dumps(list_books)))
    http_mock_sequence.append(({"status": "200"}, json.dumps(list_books_next_page)))
    for dimension in ["country", "referrer", "social_network", "source"]:
        results = {
            "reports": [
                {
                    "columnHeader": {
                        "dimensions": ["ga:pagePath", "ga:country"],
                        "metricHeader": {
                            "metricHeaderEntries": [
                                {"name": "ga:uniquePageviews", "type": "INTEGER"},
                                {"name": "ga:Pageviews", "type": "INTEGER"},
                                {"name": "ga:sessions", "type": "INTEGER"},
                            ]
                        },
                    },
                    "data": {
                        "rows": [
                            {
                                "dimensions": ["/base/path/151420", dimension + " 1"],
                                "metrics": [{"values": ["3", "4", "1"]}],
                            },
                            {
                                "dimensions": ["/base/path/151420", dimension + " 2"],
                                "metrics": [{"values": ["3", "4", "1"]}],
                            },
                            {
                                "dimensions": ["/base/path/833557", dimension + " 1"],
                                "metrics": [{"values": ["0", "0", "0"]}],  # Added a zero case for code coverage
                            },
                            {
                                "dimensions": ["/base/path/833557?fbclid=123", dimension + " 2"],
                                "metrics": [{"values": ["2", "4", "0"]}],
                            },
                        ],
                        "totals": [{"values": ["6", "9", "1"]}],
                        "rowCount": 3,
                        "minimums": [{"values": ["1", "3", "0"]}],
                        "maximums": [{"values": ["3", "4", "1"]}],
                        "isDataGolden": True,
                    },
                }
            ]
        }
        http_mock_sequence.append(({"status": "200"}, json.dumps(results)))

    return http_mock_sequence



# Copyright 2021-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Keegan Smith

import logging
import os
import re
from typing import List, Union

import pendulum
from airflow.exceptions import AirflowException
from airflow.models.taskinstance import TaskInstance
from google.cloud.bigquery import SourceFormat

from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str
from oaebu_workflows.onix_utils import collapse_subjects, onix_parser_download, onix_parser_execute
from observatory.api.client.model.dataset_release import DatasetRelease
from observatory.platform.api import make_observatory_api
from observatory.platform.airflow import AirflowConns
from observatory.platform.files import load_jsonl, save_jsonl_gz
from observatory.platform.gcs import gcs_blob_uri, gcs_upload_files, gcs_blob_name_from_path
from observatory.platform.bigquery import bq_load_table, bq_sharded_table_id, bq_create_dataset
from observatory.platform.observatory_config import CloudWorkspace
from observatory.platform.sftp import SftpFolders, make_sftp_connection
from observatory.platform.workflows.workflow import (
    SnapshotRelease,
    Workflow,
    cleanup,
    set_task_state,
    check_workflow_inputs,
)


class OnixRelease(SnapshotRelease):
    def __init__(
        self,
        *,
        dag_id: str,
        run_id: str,
        snapshot_date: pendulum.DateTime,
        onix_file_name: str,
    ):
        """Construct an OnixRelease.

        :param dag_id: The ID of the DAG
        :param run_id: The Airflow run ID
        :param snapshot_date: The date of the snapshot/release
        :param onix_file_name: The ONIX file name.
        """
        super().__init__(dag_id=dag_id, run_id=run_id, snapshot_date=snapshot_date)
        self.onix_file_name = onix_file_name
        self.download_path = os.path.join(self.download_folder, self.onix_file_name)
        self.parsed_path = os.path.join(self.transform_folder, "full.jsonl")
        self.transform_path = os.path.join(self.transform_folder, "onix.jsonl.gz")


class OnixTelescope(Workflow):
    def __init__(
        self,
        *,
        dag_id: str,
        cloud_workspace: CloudWorkspace,
        date_regex: str,
        sftp_root: str = "/",
        metadata_partner: Union[str, OaebuPartner] = "onix",
        bq_dataset_description: str = "ONIX data provided by Org",
        bq_table_description: str = None,
        api_dataset_id: str = "onix",
        observatory_api_conn_id: str = AirflowConns.OBSERVATORY_API,
        sftp_service_conn_id: str = "sftp_service",
        catchup: bool = False,
        schedule: str = "@weekly",
        start_date: pendulum.DateTime = pendulum.datetime(2021, 3, 28),
    ):
        """Construct an OnixTelescope instance.
        :param dag_id: The ID of the DAG
        :param cloud_workspace: The CloudWorkspace object for this DAG
        :param sftp_root: The working root of the SFTP server, passed to the SftoFolders class
        :param metadata_partner: The metadata partner name
        :param date_regex: Regular expression for extracting a date string from an ONIX file name
        :param bq_dataset_description: Description for the BigQuery dataset
        :param bq_table_description: Description for the biguery table
        :param api_dataset_id: The ID to store the dataset release in the API
        :param observatory_api_conn_id: Airflow connection ID for the overvatory API
        :param sftp_service_conn_id: Airflow connection ID for the SFTP service
        :param catchup: Whether to catchup the DAG or not
        :param schedule: The schedule interval of the DAG
        :param start_date: The start date of the DAG
        """
        super().__init__(
            dag_id,
            start_date,
            schedule,
            catchup=catchup,
            airflow_conns=[observatory_api_conn_id, sftp_service_conn_id],
            tags=["oaebu"],
        )
        self.dag_id = dag_id
        self.cloud_workspace = cloud_workspace
        self.sftp_root = sftp_root
        self.metadata_partner = partner_from_str(metadata_partner, metadata_partner=True)
        self.date_regex = date_regex
        self.bq_dataset_description = bq_dataset_description
        self.bq_table_description = bq_table_description
        self.api_dataset_id = api_dataset_id
        self.observatory_api_conn_id = observatory_api_conn_id
        self.sftp_service_conn_id = sftp_service_conn_id
        self.sftp_folders = SftpFolders(dag_id, sftp_conn_id=sftp_service_conn_id, sftp_root=sftp_root)

        check_workflow_inputs(self)

        self.add_setup_task(self.check_dependencies)
        self.add_setup_task(self.list_release_info)
        self.add_task(self.move_files_to_in_progress)
        self.add_task(self.download)
        self.add_task(self.upload_downloaded)
        self.add_task(self.transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.move_files_to_finished)
        self.add_task(self.add_new_dataset_releases)
        self.add_task(self.cleanup)

    def list_release_info(self, **kwargs):
        """Lists all ONIX releases and publishes their file names as an XCom.

        :param kwargs: the context passed from the BranchPythonOperator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for the keyword arguments that can be passed
        :return: the identifier of the task to execute next.
        """

        # List release dates
        release_info = []
        with make_sftp_connection(self.sftp_service_conn_id) as sftp:
            files = sftp.listdir(self.sftp_folders.upload)
            for file_name in files:
                if re.match(r"^.*\.(onx|xml)$", file_name):
                    try:
                        date_str = re.search(self.date_regex, file_name).group(0)
                    except AttributeError:
                        msg = f"Could not find date with pattern `{self.date_regex}` in file name {file_name}"
                        logging.error(msg)
                        raise AirflowException(msg)
                    release_info.append({"release_date": date_str, "file_name": file_name})

        # Publish XCom
        continue_dag = len(release_info)
        if continue_dag:
            ti: TaskInstance = kwargs["ti"]
            execution_date = kwargs["execution_date"]
            ti.xcom_push(OnixTelescope.RELEASE_INFO, release_info, execution_date)

        return continue_dag

    def make_release(self, **kwargs) -> List[OnixRelease]:
        """Make release instances. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :return: a list of Onix release instances.
        """

        ti: TaskInstance = kwargs["ti"]
        records = ti.xcom_pull(
            key=OnixTelescope.RELEASE_INFO, task_ids=self.list_release_info.__name__, include_prior_dates=False
        )
        releases = []
        for record in records:
            onix_file_name = record["file_name"]
            releases.append(
                OnixRelease(
                    dag_id=self.dag_id,
                    run_id=kwargs["run_id"],
                    snapshot_date=pendulum.parse(record["release_date"]),
                    onix_file_name=onix_file_name,
                )
            )
        return releases

    def move_files_to_in_progress(self, releases: List[OnixRelease], **kwargs):
        """Move ONIX files to SFTP in-progress folder.
        :param releases: a list of Onix release instances"""
        self.sftp_folders.move_files_to_in_progress([release.onix_file_name for release in releases])

    def download(self, releases: List[OnixRelease], **kwargs):
        """Task to download the ONIX releases."""
        with make_sftp_connection(self.sftp_service_conn_id) as sftp:
            for release in releases:
                in_progress_file = os.path.join(self.sftp_folders.in_progress, release.onix_file_name)
                sftp.get(in_progress_file, localpath=release.download_path)

    def upload_downloaded(self, releases: List[OnixRelease], **kwargs):
        """Uploads the downloaded onix file to GCS"""
        for release in releases:
            success = gcs_upload_files(
                bucket_name=self.cloud_workspace.download_bucket, file_paths=[release.download_path]
            )
            set_task_state(success, kwargs["ti"].task_id, release=release)

    def transform(self, releases: List[OnixRelease], **kwargs):
        """Task to transform the ONIX releases."""

        success, parser_path = onix_parser_download()
        set_task_state(success, kwargs["ti"].task_id)
        for release in releases:
            onix_parser_execute(
                parser_path=parser_path, input_dir=release.download_folder, output_dir=release.transform_folder
            )
            onix = collapse_subjects(load_jsonl(release.parsed_path))
            save_jsonl_gz(release.transform_path, onix)

    def upload_transformed(self, releases: List[OnixRelease], **kwargs):
        """Uploads the transformed file to GCS"""
        for release in releases:
            success = gcs_upload_files(
                bucket_name=self.cloud_workspace.transform_bucket, file_paths=[release.transform_path]
            )
            set_task_state(success, kwargs["ti"].task_id, release=release)

    def bq_load(self, releases: List[OnixRelease], **kwargs):
        """Task to load each transformed release to BigQuery."""
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.metadata_partner.bq_dataset_id,
            location=self.cloud_workspace.data_location,
            description=self.bq_dataset_description,
        )
        # Load each transformed release
        for release in releases:
            table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.metadata_partner.bq_dataset_id,
                self.metadata_partner.bq_table_name,
                release.snapshot_date,
            )
            uri = gcs_blob_uri(self.cloud_workspace.transform_bucket, gcs_blob_name_from_path(release.transform_path))
            state = bq_load_table(
                uri=uri,
                table_id=table_id,
                schema_file_path=self.metadata_partner.schema_path,
                source_format=SourceFormat.NEWLINE_DELIMITED_JSON,
                table_description=self.bq_table_description,
            )
            set_task_state(state, kwargs["ti"].task_id, release=release)

    def move_files_to_finished(self, releases: List[OnixRelease], **kwargs):
        """Move ONIX files to SFTP finished folder."""
        self.sftp_folders.move_files_to_finished([release.onix_file_name for release in releases])

    def add_new_dataset_releases(self, releases: List[OnixRelease], **kwargs) -> None:
        """Adds release information to API."""
        api = make_observatory_api(observatory_api_conn_id=self.observatory_api_conn_id)
        for release in releases:
            dataset_release = DatasetRelease(
                dag_id=self.dag_id,
                dataset_id=self.api_dataset_id,
                dag_run_id=release.run_id,
                snapshot_date=release.snapshot_date,
                data_interval_start=kwargs["data_interval_start"],
                data_interval_end=kwargs["data_interval_end"],
            )
            api.post_dataset_release(dataset_release)

    def cleanup(self, releases: List[OnixRelease], **kwargs) -> None:
        """Delete all files, folders and XComs associated with this release."""
        for release in releases:
            cleanup(
                dag_id=self.dag_id, execution_date=kwargs["execution_date"], workflow_folder=release.workflow_folder
            )



# Copyright 2022-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Keegan Smith

import os

import pendulum
import vcr
from airflow.utils.state import State
from airflow.models.connection import Connection

from oaebu_workflows.config import test_fixtures_folder
from oaebu_workflows.oaebu_partners import partner_from_str
from oaebu_workflows.irus_fulcrum_telescope.irus_fulcrum_telescope import (
    IrusFulcrumTelescope,
    download_fulcrum_month_data,
    transform_fulcrum_data,
)
from observatory.platform.files import load_jsonl
from observatory.platform.observatory_environment import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    find_free_port,
    load_and_parse_json,
)
from observatory.platform.api import get_dataset_releases
from observatory.platform.gcs import gcs_blob_name_from_path
from observatory.platform.bigquery import bq_table_id
from observatory.platform.observatory_config import Workflow

FAKE_PUBLISHERS = ["Fake Publisher 1", "Fake Publisher 2", "Fake Publisher 3"]


class TestIrusFulcrumTelescope(ObservatoryTestCase):
    """Tests for the Fulcrum telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")

        # Fixtures
        fixtures_folder = test_fixtures_folder(workflow_module="irus_fulcrum_telescope")
        self.download_cassette = os.path.join(fixtures_folder, "fulcrum_download_cassette.yaml")
        self.test_table = os.path.join(fixtures_folder, "test_final_table.json")
        self.test_totals_download = os.path.join(fixtures_folder, "test_totals_download.jsonl")
        self.test_country_download = os.path.join(fixtures_folder, "test_country_download.jsonl")
        self.test_transform = os.path.join(fixtures_folder, "test_transform.jsonl")

    def test_dag_structure(self):
        """Test that the ONIX DAG has the correct structure and raises errors when necessary"""
        dag = IrusFulcrumTelescope(
            dag_id="fulcrum_test", cloud_workspace=self.fake_cloud_workspace, publishers=FAKE_PUBLISHERS
        ).make_dag()

        self.assert_dag_structure(
            {
                "check_dependencies": ["download"],
                "download": ["upload_downloaded"],
                "upload_downloaded": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["add_new_dataset_releases"],
                "add_new_dataset_releases": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the DAG can be loaded from a DAG bag."""
        env = ObservatoryEnvironment(
            workflows=[
                Workflow(
                    dag_id="fulcrum_test",
                    name="Fulcrum Telescope",
                    class_name="oaebu_workflows.irus_fulcrum_telescope.irus_fulcrum_telescope.IrusFulcrumTelescope",
                    cloud_workspace=self.fake_cloud_workspace,
                    kwargs=dict(publishers=[FAKE_PUBLISHERS]),
                )
            ]
        )
        with env.create():
            self.assert_dag_load_from_config("fulcrum_test")

    def test_telescope(self):
        """Test the Fulcrum telescope end to end."""
        # Setup Observatory environment
        env = ObservatoryEnvironment(
            self.project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        # Create the Observatory environment and run tests
        with env.create():
            # Setup Telescope
            execution_date = pendulum.datetime(year=2022, month=4, day=7)
            partner = partner_from_str("irus_fulcrum")
            partner.bq_dataset_id = env.add_dataset()
            telescope = IrusFulcrumTelescope(
                dag_id="fulcrum_test",
                cloud_workspace=env.cloud_workspace,
                publishers=FAKE_PUBLISHERS,
                data_partner=partner,
            )
            dag = telescope.make_dag()
            env.add_connection(Connection(conn_id=telescope.irus_oapen_api_conn_id, uri=f"http://fake_api_login:@"))

            # Add the fake requestor ID as a connection
            with env.create_dag_run(dag, execution_date):
                # Test that all dependencies are specified: no error should be thrown
                ti = env.run_task(telescope.check_dependencies.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test download
                fulcrum_vcr = vcr.VCR(record_mode="none")
                with fulcrum_vcr.use_cassette(self.download_cassette):
                    ti = env.run_task(telescope.download.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)

                # Test upload downloaded
                ti = env.run_task(telescope.upload_downloaded.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test transform
                ti = env.run_task(telescope.transform.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test upload to cloud storage
                ti = env.run_task(telescope.upload_transformed.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test load into BigQuery
                ti = env.run_task(telescope.bq_load.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                ### Make assertions ##

                # Create the release
                release = telescope.make_release(
                    run_id=env.dag_run.run_id,
                    data_interval_start=pendulum.parse(str(env.dag_run.data_interval_start)),
                    data_interval_end=pendulum.parse(str(env.dag_run.data_interval_end)),
                )

                # Downloaded files
                self.assert_file_integrity(release.download_totals_path, "95b7dceb", "gzip_crc")
                self.assert_file_integrity(release.download_country_path, "0a713d03", "gzip_crc")

                # Uploaded download blob
                self.assert_blob_integrity(
                    env.download_bucket,
                    gcs_blob_name_from_path(release.download_totals_path),
                    release.download_totals_path,
                )
                self.assert_blob_integrity(
                    env.download_bucket,
                    gcs_blob_name_from_path(release.download_country_path),
                    release.download_country_path,
                )

                # Transformed file
                self.assert_file_integrity(release.transform_path, "40a25e4e", "gzip_crc")

                # Uploaded transform blob
                self.assert_blob_integrity(
                    env.transform_bucket, gcs_blob_name_from_path(release.transform_path), release.transform_path
                )

                # Uploaded table
                table_id = bq_table_id(
                    telescope.cloud_workspace.project_id,
                    telescope.data_partner.bq_dataset_id,
                    telescope.data_partner.bq_table_name,
                )
                self.assert_table_integrity(table_id, expected_rows=3)
                self.assert_table_content(
                    table_id,
                    load_and_parse_json(self.test_table, date_fields="release_date"),
                    primary_key="proprietary_id",
                )

                # Add_dataset_release_task
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 0)
                ti = env.run_task(telescope.add_new_dataset_releases.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 1)

                # Test cleanup
                ti = env.run_task(telescope.cleanup.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_cleanup(release.workflow_folder)

    def test_download_fulcrum_month_data(self):
        """Tests the download_fuclrum_month_data function"""
        vcr_ = vcr.VCR(record_mode="none")
        with vcr_.use_cassette(self.download_cassette):
            actual_totals, actual_country = download_fulcrum_month_data(
                download_month=pendulum.datetime(year=2022, month=4, day=15),
                requestor_id="fake_api_login",
                num_retries=0,
            )
        expected_totals = load_jsonl(self.test_totals_download)
        expected_country = load_jsonl(self.test_country_download)

        # Make list order deterministic before testing
        actual_totals = [dict(sorted(d.items())) for d in actual_totals]
        actual_country = [dict(sorted(d.items())) for d in actual_country]
        expected_totals = [dict(sorted(d.items())) for d in expected_totals]
        expected_country = [dict(sorted(d.items())) for d in expected_country]
        self.assertListEqual(actual_totals, expected_totals)
        self.assertListEqual(actual_country, expected_country)

    def test_transform_fulcrum_data(self):
        """Tests the transform_fulcrum_data function"""
        totals = load_jsonl(self.test_totals_download)
        country = load_jsonl(self.test_country_download)
        actual_transform = transform_fulcrum_data(totals_data=totals, country_data=country, publishers=FAKE_PUBLISHERS)
        expected_transform = load_jsonl(self.test_transform)

        # Make list order deterministic before testing
        actual_transform = [dict(sorted(d.items())) for d in actual_transform]
        expected_transform = [dict(sorted(d.items())) for d in expected_transform]
        self.assertListEqual(actual_transform, expected_transform)



# Copyright 2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Keegan Smith

import os
from tempfile import TemporaryDirectory

import pendulum
import vcr
from airflow.utils.state import State

from oaebu_workflows.oaebu_partners import partner_from_str
from oaebu_workflows.thoth_telescope.thoth_telescope import (
    ThothTelescope,
    thoth_download_onix,
    DEFAULT_HOST_NAME,
)
from oaebu_workflows.config import test_fixtures_folder
from observatory.platform.api import get_dataset_releases
from observatory.platform.bigquery import bq_sharded_table_id
from observatory.platform.gcs import gcs_blob_name_from_path
from observatory.platform.utils.url_utils import retry_get_url
from observatory.platform.observatory_config import Workflow
from observatory.platform.observatory_environment import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    find_free_port,
    load_and_parse_json,
)


FAKE_PUBLISHER_ID = "fake_publisher_id"


class TestThothTelescope(ObservatoryTestCase):
    """Tests for the Thoth telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """
        super(TestThothTelescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")

        # Fixtures
        fixtures_folder = test_fixtures_folder(workflow_module="thoth_telescope")
        self.download_cassette = os.path.join(fixtures_folder, "thoth_download_cassette.yaml")
        self.test_table = os.path.join(fixtures_folder, "test_table.json")

    def test_dag_structure(self):
        """Test that the ONIX DAG has the correct structure."""

        dag = ThothTelescope(
            dag_id="thoth_telescope_test",
            cloud_workspace=self.fake_cloud_workspace,
            publisher_id=FAKE_PUBLISHER_ID,
            format_specification="onix_3.0::jstor",
        ).make_dag()

        self.assert_dag_structure(
            {
                "check_dependencies": ["download"],
                "download": ["upload_downloaded"],
                "upload_downloaded": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["add_new_dataset_releases"],
                "add_new_dataset_releases": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the DAG can be loaded from a DAG bag."""
        env = ObservatoryEnvironment(
            workflows=[
                Workflow(
                    dag_id="thoth_telescope_test",
                    name="Thoth Telescope",
                    class_name="oaebu_workflows.thoth_telescope.thoth_telescope.ThothTelescope",
                    cloud_workspace=self.fake_cloud_workspace,
                    kwargs=dict(publisher_id=FAKE_PUBLISHER_ID, format_specification="onix::oapen"),
                )
            ],
        )
        with env.create():
            self.assert_dag_load_from_config("thoth_telescope_test")

        # Error should be raised for no publisher_id
        env.workflows[0].kwargs = {}
        with env.create():
            with self.assertRaises(AssertionError) as cm:
                self.assert_dag_load_from_config("onix_workflow_test_dag_load")
            msg = cm.exception.args[0]
            self.assertIn("missing 2 required keyword-only arguments", msg)
            self.assertIn("publisher_id", msg)
            self.assertIn("format_specification", msg)

    def test_telescope(self):
        """Test the Thoth telescope end to end."""
        env = ObservatoryEnvironment(
            self.project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )

        # Create the Observatory environment and run tests
        with env.create():
            # Setup Telescope
            execution_date = pendulum.datetime(year=2022, month=12, day=1)
            metadata_partner = partner_from_str("thoth", metadata_partner=True)
            metadata_partner.bq_dataset_id = env.add_dataset()
            telescope = ThothTelescope(
                dag_id="thoth_telescope_test",
                cloud_workspace=env.cloud_workspace,
                format_specification="onix_3.0::oapen",
                publisher_id=FAKE_PUBLISHER_ID,
                metadata_partner=metadata_partner,
            )
            dag = telescope.make_dag()

            with env.create_dag_run(dag, execution_date):
                ti = env.run_task(telescope.check_dependencies.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                thoth_vcr = vcr.VCR(record_mode="none")
                with thoth_vcr.use_cassette(self.download_cassette):
                    ti = env.run_task(telescope.download.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                ti = env.run_task(telescope.upload_downloaded.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                ti = env.run_task(telescope.transform.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                ti = env.run_task(telescope.upload_transformed.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                ti = env.run_task(telescope.bq_load.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                ### Make assertions ###

                # Make the release
                release = telescope.make_release(
                    run_id=env.dag_run.run_id, data_interval_end=pendulum.parse(str(env.dag_run.data_interval_end))
                )

                # Downloaded file
                self.assert_file_integrity(release.download_path, "6d0f31c315dab144054e2fde9ad7f8ab", "md5")

                # Uploaded download blob
                self.assert_blob_integrity(
                    env.download_bucket, gcs_blob_name_from_path(release.download_path), release.download_path
                )

                # Transformed file
                self.assert_file_integrity(release.transform_path, "1b12c3a9", "gzip_crc")

                # Uploaded transform blob
                self.assert_blob_integrity(
                    env.transform_bucket, gcs_blob_name_from_path(release.transform_path), release.transform_path
                )

                # Uploaded table
                table_id = bq_sharded_table_id(
                    telescope.cloud_workspace.project_id,
                    telescope.metadata_partner.bq_dataset_id,
                    telescope.metadata_partner.bq_table_name,
                    release.snapshot_date,
                )
                self.assert_table_integrity(table_id, expected_rows=2)
                self.assert_table_content(table_id, load_and_parse_json(self.test_table), primary_key="ISBN13")

                # add_dataset_release_task
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 0)
                ti = env.run_task(telescope.add_new_dataset_releases.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 1)

                # Test cleanup
                ti = env.run_task(telescope.cleanup.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_cleanup(release.workflow_folder)

    # Function tests
    def test_download_onix(self):
        """
        Tests the download_onix function.
        Will test that the function works as expected using proxy HTTP requests
        """
        with TemporaryDirectory() as tempdir:
            thoth_vcr = vcr.VCR(record_mode="none")
            with thoth_vcr.use_cassette(self.download_cassette):
                thoth_download_onix(
                    FAKE_PUBLISHER_ID,
                    download_path=os.path.join(tempdir, "fake_download.xml"),
                    num_retries=0,
                    format_spec="onix_3.0::oapen",
                )
            self.assert_file_integrity(
                os.path.join(tempdir, "fake_download.xml"), "6d0f31c315dab144054e2fde9ad7f8ab", "md5"
            )

    def test_thoth_api(self):
        """Tests that HTTP requests to the thoth API are successful"""
        base_response = retry_get_url(DEFAULT_HOST_NAME, num_retries=2)
        format_response = retry_get_url(f"{DEFAULT_HOST_NAME}/specifications/onix_3.0::oapen", num_retries=2)
        self.assertEqual(base_response.status_code, 200)
        self.assertEqual(format_response.status_code, 200)



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, Keegan Smith

import base64
import json
import os
from unittest.mock import Mock, patch

import httpretty
import pendulum
from airflow.exceptions import AirflowException
from airflow.models.connection import Connection
from airflow.utils.state import State
from click.testing import CliRunner
from googleapiclient.discovery import build
from googleapiclient.http import HttpMockSequence

from oaebu_workflows.config import test_fixtures_folder
from oaebu_workflows.oaebu_partners import partner_from_str
from oaebu_workflows.jstor_telescope.jstor_telescope import (
    JstorRelease,
    JstorTelescope,
    JstorPublishersAPI,
    JstorCollectionsAPI,
    make_jstor_api,
)
from observatory.platform.observatory_environment import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    find_free_port,
    load_and_parse_json,
)
from observatory.platform.observatory_config import Workflow
from observatory.platform.gcs import gcs_blob_name_from_path
from observatory.platform.bigquery import bq_table_id
from observatory.platform.api import get_dataset_releases


def dummy_gmail_connection() -> Connection:
    return Connection(
        conn_id="gmail_api",
        uri="google-cloud-platform://?token=123&refresh_token=123"
        "&client_id=123.apps.googleusercontent.com&client_secret=123",
    )


class TestTelescopeSetup(ObservatoryTestCase):
    def __init__(self, *args, **kwargs):
        super(TestTelescopeSetup, self).__init__(*args, **kwargs)
        self.entity_id = "anupress"

    def test_dag_structure(self):
        """Test that the Jstor DAG has the correct structure."""
        env = ObservatoryEnvironment()
        with env.create():
            env.add_connection(dummy_gmail_connection())
            for entity_type in ["publisher", "collection"]:
                dag = JstorTelescope(
                    "jstor",
                    cloud_workspace=self.fake_cloud_workspace,
                    entity_id=self.entity_id,
                    entity_type=entity_type,
                ).make_dag()
                self.assert_dag_structure(
                    {
                        "check_dependencies": ["list_reports"],
                        "list_reports": ["download_reports"],
                        "download_reports": ["upload_downloaded"],
                        "upload_downloaded": ["transform"],
                        "transform": ["upload_transformed"],
                        "upload_transformed": ["bq_load"],
                        "bq_load": ["add_new_dataset_releases"],
                        "add_new_dataset_releases": ["cleanup"],
                        "cleanup": [],
                    },
                    dag,
                )

    def test_dag_load(self):
        """Test that the Jstor DAG can be loaded from a DAG bag."""
        for entity_type in ["publisher", "collection"]:
            env = ObservatoryEnvironment(
                workflows=[
                    Workflow(
                        dag_id="jstor_test_telescope",
                        name="My JSTOR Workflow",
                        class_name="oaebu_workflows.jstor_telescope.jstor_telescope.JstorTelescope",
                        cloud_workspace=self.fake_cloud_workspace,
                        kwargs=dict(entity_id=self.entity_id, entity_type=entity_type),
                    )
                ],
            )

            with env.create():
                env.add_connection(dummy_gmail_connection())
                self.assert_dag_load_from_config("jstor_test_telescope")

    def test_entity_type(self):
        env = ObservatoryEnvironment()
        with env.create():
            env.add_connection(dummy_gmail_connection())
            for entity_type in ["collection", "publisher"]:
                make_jstor_api(entity_type, self.entity_id)
            with self.assertRaisesRegex(AirflowException, "Entity type must be"):
                make_jstor_api("invalid", self.entity_id)


class TestJstorTelescopePublisher(ObservatoryTestCase):
    """Tests for the Jstor telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """
        super(TestJstorTelescopePublisher, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.entity_id = "anupress"

        self.release_date = pendulum.parse("20220701").end_of("month")
        fixtures_folder = test_fixtures_folder(workflow_module="jstor_telescope")
        self.country_report = {
            "path": os.path.join(fixtures_folder, "country_20220801.tsv"),
            "url": "https://www.jstor.org/admin/reports/download/249192019",
            "headers": {
                "Content-Disposition": f"attachment; filename=PUB_{self.entity_id}_PUBBCU_"
                f'{self.release_date.strftime("%Y%m%d")}.tsv'
            },
            "download_hash": "9330cc71f8228838ac84abb33cedb3b8",
            "transform_hash": "5a72fe64",
            "table_rows": 10,
        }
        self.institution_report = {
            "path": os.path.join(fixtures_folder, "institution_20220801.tsv"),
            "url": "https://www.jstor.org/admin/reports/download/129518301",
            "headers": {
                "Content-Disposition": f"attachment; filename=PUB_{self.entity_id}_PUBBIU_"
                f'{self.release_date.strftime("%Y%m%d")}.tsv'
            },
            "download_hash": "1c78c316766a3f7306d6c19440250484",
            "transform_hash": "f339ea81",
            "table_rows": 3,
        }
        self.wrong_publisher_report = {
            "path": os.path.join(fixtures_folder, "institution_20220801.tsv"),  # has to be valid path, but is not used
            "url": "https://www.jstor.org/admin/reports/download/12345",
            "headers": {
                "Content-Disposition": f"attachment; filename=PUB_publisher_PUBBIU_"
                f'{self.release_date.strftime("%Y%m%d")}.tsv'
            },
        }

    @patch("oaebu_workflows.jstor_telescope.jstor_telescope.build")
    @patch("oaebu_workflows.jstor_telescope.jstor_telescope.Credentials")
    def test_telescope_publisher(self, mock_account_credentials, mock_build):
        """Test the Jstor telescope end to end."""

        mock_account_credentials.from_json_keyfile_dict.return_value = ""

        http = HttpMockSequence(
            publisher_http_mock_sequence(
                self.country_report["url"], self.institution_report["url"], self.wrong_publisher_report["url"]
            )
        )
        mock_build.return_value = build("gmail", "v1", http=http)

        # Setup Observatory environment
        env = ObservatoryEnvironment(
            self.project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        dataset_id = env.add_dataset()

        # Create the Observatory environment and run tests
        with env.create(task_logging=True):
            # Add gmail connection
            env.add_connection(dummy_gmail_connection())

            # Setup Telescope
            execution_date = pendulum.datetime(year=2020, month=11, day=1)
            country_partner = partner_from_str("jstor_country")
            country_partner.bq_dataset_id = dataset_id
            institution_partner = partner_from_str("jstor_institution")
            institution_partner.bq_dataset_id = dataset_id
            telescope = JstorTelescope(
                dag_id="jstor_test_telescope",
                cloud_workspace=env.cloud_workspace,
                entity_id=self.entity_id,
                entity_type="publisher",
                country_partner=country_partner,
                institution_partner=institution_partner,
            )
            dag = telescope.make_dag()

            # Begin DAG run
            with env.create_dag_run(dag, execution_date):
                # Test that all dependencies are specified: no error should be thrown
                ti = env.run_task(telescope.check_dependencies.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test list releases task with files available
                with httpretty.enabled():
                    for report in [self.country_report, self.institution_report, self.wrong_publisher_report]:
                        self.setup_mock_file_download(
                            report["url"], report["path"], headers=report["headers"], method=httpretty.HEAD
                        )
                    ti = env.run_task(telescope.list_reports.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                available_reports = ti.xcom_pull(
                    key=JstorTelescope.REPORTS_INFO, task_ids=telescope.list_reports.__name__, include_prior_dates=False
                )
                self.assertIsInstance(available_reports, list)
                expected_reports_info = [
                    {"type": "country", "url": self.country_report["url"], "id": "1788ec9e91f3de62"},
                    {"type": "institution", "url": self.institution_report["url"], "id": "1788ebe4ecbab055"},
                ]
                self.assertListEqual(expected_reports_info, available_reports)

                # Test download_reports task
                with httpretty.enabled():
                    for report in [self.country_report, self.institution_report]:
                        self.setup_mock_file_download(report["url"], report["path"], headers=report["headers"])
                    ti = env.run_task(telescope.download_reports.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)

                # use release info for other tasks
                available_releases = ti.xcom_pull(
                    key=JstorTelescope.RELEASE_INFO,
                    task_ids=telescope.download_reports.__name__,
                    include_prior_dates=False,
                )
                self.assertIsInstance(available_releases, dict)
                self.assertEqual(1, len(available_releases))
                for release_date, reports in available_releases.items():
                    self.assertEqual(self.release_date.date(), pendulum.parse(release_date).date())
                    self.assertIsInstance(reports, list)
                    self.assertListEqual(expected_reports_info, reports)
                release = JstorRelease(
                    dag_id=telescope.dag_id,
                    run_id=env.dag_run.run_id,
                    data_interval_start=pendulum.parse(release_date).start_of("month"),
                    data_interval_end=pendulum.parse(release_date).add(days=1).start_of("month"),
                    partition_date=pendulum.parse(release_date),
                    reports=reports,
                )

                self.assertTrue(os.path.exists(release.download_country_path))
                self.assertTrue(os.path.exists(release.download_institution_path))
                self.assert_file_integrity(release.download_country_path, self.country_report["download_hash"], "md5")
                self.assert_file_integrity(
                    release.download_institution_path, self.institution_report["download_hash"], "md5"
                )
                # Test that file uploaded
                ti = env.run_task(telescope.upload_downloaded.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_blob_integrity(
                    env.download_bucket,
                    gcs_blob_name_from_path(release.download_country_path),
                    release.download_country_path,
                )
                self.assert_blob_integrity(
                    env.download_bucket,
                    gcs_blob_name_from_path(release.download_institution_path),
                    release.download_institution_path,
                )

                # Test that file transformed
                ti = env.run_task(telescope.transform.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assertTrue(os.path.exists(release.transform_country_path))
                self.assertTrue(os.path.exists(release.transform_institution_path))
                self.assert_file_integrity(
                    release.transform_country_path, self.country_report["transform_hash"], "gzip_crc"
                )
                self.assert_file_integrity(
                    release.transform_institution_path, self.institution_report["transform_hash"], "gzip_crc"
                )

                # Test that transformed file uploaded
                ti = env.run_task(telescope.upload_transformed.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_blob_integrity(
                    env.transform_bucket,
                    gcs_blob_name_from_path(release.transform_country_path),
                    release.transform_country_path,
                )
                self.assert_blob_integrity(
                    env.transform_bucket,
                    gcs_blob_name_from_path(release.transform_institution_path),
                    release.transform_institution_path,
                )

                # Test that data loaded into BigQuery
                ti = env.run_task(telescope.bq_load.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                country_table_id = bq_table_id(
                    telescope.cloud_workspace.project_id,
                    telescope.country_partner.bq_dataset_id,
                    telescope.country_partner.bq_table_name,
                )
                institution_table_id = bq_table_id(
                    telescope.cloud_workspace.project_id,
                    telescope.institution_partner.bq_dataset_id,
                    telescope.institution_partner.bq_table_name,
                )
                self.assert_table_integrity(country_table_id, self.country_report["table_rows"])
                self.assert_table_integrity(institution_table_id, self.institution_report["table_rows"])

                # Add_dataset_release_task
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 0)
                ti = env.run_task(telescope.add_new_dataset_releases.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 1)

                # Test that all telescope data deleted
                ti = env.run_task(telescope.cleanup.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_cleanup(release.workflow_folder)

    def test_get_release_date(self):
        """Test that the get_release_date returns the correct release date and raises an exception when dates are
        incorrect"""
        with CliRunner().isolated_filesystem():
            # Test reports in new format with header
            new_report_content = (
                "Report_Name\tBook Usage by Country\nReport_ID\tPUB_BCU\nReport_Description\t"
                "Usage of your books on JSTOR by country.\nPublisher_Name\tPublisher 1\nReporting_Period\t"
                "{start_date} to {end_date}\nCreated\t2021-10-01\nCreated_By\tJSTOR"
            )
            old_report_content = (
                "Book Title\tUsage Month\nAddress source war after\t{start_date}\nNote spend " "government\t{end_date}"
            )
            reports = [
                {"file": "new_success.tsv", "start": "2020-01-01", "end": "2020-01-31"},
                {"file": "new_fail.tsv", "start": "2020-01-01", "end": "2020-02-01"},
                {"file": "old_success.tsv", "start": "2020-01", "end": "2020-01"},
                {"file": "old_fail.tsv", "start": "2020-01", "end": "2020-02"},
            ]

            for report in reports:
                with open(report["file"], "w") as f:
                    if report == reports[0] or report == reports[1]:
                        f.write(new_report_content.format(start_date=report["start"], end_date=report["end"]))
                    else:
                        f.write(old_report_content.format(start_date=report["start"], end_date=report["end"]))

            # Test new report is successful
            api = JstorPublishersAPI(Mock(), self.entity_id)
            start_date, end_date = api.get_release_date(reports[0]["file"])
            self.assertEqual(pendulum.parse(reports[0]["start"]), start_date)
            self.assertEqual(pendulum.parse(reports[0]["end"]), end_date)

            # Test new report fails
            with self.assertRaises(AirflowException):
                api.get_release_date(reports[1]["file"])

            # Test old report is successful
            start_date, end_date = api.get_release_date(reports[2]["file"])
            self.assertEqual(pendulum.parse(reports[2]["start"]).start_of("month").start_of("day"), start_date)
            self.assertEqual(pendulum.parse(reports[2]["end"]).end_of("month").start_of("day"), end_date)

            # Test old report fails
            with self.assertRaises(AirflowException):
                api.get_release_date(reports[3]["file"])


class TestJstorTelescopeCollection(ObservatoryTestCase):
    """Tests for the Jstor telescope"""

    def __init__(self, *args, **kwargs):
        super(TestJstorTelescopeCollection, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.entity_id = "BTAA"
        self.release_date = pendulum.parse("20230901").end_of("month")

        fixtures_folder = test_fixtures_folder(workflow_module="jstor_telescope")
        self.country_report = {
            "path": os.path.join(fixtures_folder, "collection_country.json"),
            "table": os.path.join(fixtures_folder, "collection_country_table.json"),
            "url": "https://www.jstor.org/admin/reports/download/249192019",
            "download_hash": "8b03d0353ac258bfca48c9347dab76af",
            "transform_hash": "885bf71d",
            "table_rows": 3,
        }
        self.institution_report = {
            "path": os.path.join(fixtures_folder, "collection_institution.json"),
            "table": os.path.join(fixtures_folder, "collection_institution_table.json"),
            "url": "https://www.jstor.org/admin/reports/download/129518301",
            "download_hash": "1507899df3b3937d6fc20cbe007c882f",
            "transform_hash": "6ff261d4",
            "table_rows": 6,
        }

    @patch("oaebu_workflows.jstor_telescope.jstor_telescope.build")
    @patch("oaebu_workflows.jstor_telescope.jstor_telescope.Credentials")
    def test_telescope_collection(self, mock_account_credentials, mock_build):
        """Test the Jstor telescope end to end."""

        mock_account_credentials.from_json_keyfile_dict.return_value = ""
        http = HttpMockSequence(
            collection_http_mock_sequence(self.country_report["path"], self.institution_report["path"])
        )
        mock_build.return_value = build("gmail", "v1", http=http)

        # Setup Observatory environment
        env = ObservatoryEnvironment(
            self.project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        dataset_id = env.add_dataset()

        # Create the Observatory environment and run tests
        with env.create(task_logging=True):
            # Add gmail connection
            env.add_connection(dummy_gmail_connection())

            # Setup Telescope
            execution_date = pendulum.datetime(year=2023, month=10, day=4)
            country_partner = partner_from_str("jstor_country_collection")
            country_partner.bq_dataset_id = dataset_id
            institution_partner = partner_from_str("jstor_institution_collection")
            institution_partner.bq_dataset_id = dataset_id
            telescope = JstorTelescope(
                dag_id="jstor_test_telescope",
                cloud_workspace=env.cloud_workspace,
                entity_id=self.entity_id,
                entity_type="collection",
                country_partner=country_partner,
                institution_partner=institution_partner,
            )
            dag = telescope.make_dag()

            # Begin DAG run
            with env.create_dag_run(dag, execution_date):
                # Test that all dependencies are specified: no error should be thrown
                ti = env.run_task(telescope.check_dependencies.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test list releases task with files available
                ti = env.run_task(telescope.list_reports.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                available_reports = ti.xcom_pull(
                    key=JstorTelescope.REPORTS_INFO, task_ids=telescope.list_reports.__name__, include_prior_dates=False
                )
                self.assertIsInstance(available_reports, list)
                expected_reports_info = [
                    {"type": "country", "attachment_id": "2", "id": "18af0b40b64fe408"},
                    {"type": "institution", "attachment_id": "3", "id": "18af0b40b64fe408"},
                ]
                self.assertListEqual(expected_reports_info, available_reports)

                # Test download_reports task
                ti = env.run_task(telescope.download_reports.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # use release info for other tasks
                available_releases = ti.xcom_pull(
                    key=JstorTelescope.RELEASE_INFO,
                    task_ids=telescope.download_reports.__name__,
                    include_prior_dates=False,
                )
                self.assertIsInstance(available_releases, dict)
                self.assertEqual(1, len(available_releases))
                for release_date, reports in available_releases.items():
                    self.assertEqual(self.release_date.date(), pendulum.parse(release_date).date())
                    self.assertIsInstance(reports, list)
                    self.assertListEqual(expected_reports_info, reports)

                release = JstorRelease(
                    dag_id=telescope.dag_id,
                    run_id=env.dag_run.run_id,
                    data_interval_start=pendulum.parse(release_date).start_of("month"),
                    data_interval_end=pendulum.parse(release_date).add(days=1).start_of("month"),
                    partition_date=pendulum.parse(release_date),
                    reports=reports,
                )

                # Test that files download
                self.assertTrue(os.path.exists(release.download_country_path))
                self.assertTrue(os.path.exists(release.download_institution_path))
                self.assert_file_integrity(release.download_country_path, self.country_report["download_hash"], "md5")
                self.assert_file_integrity(
                    release.download_institution_path, self.institution_report["download_hash"], "md5"
                )

                # Test that file uploaded
                ti = env.run_task(telescope.upload_downloaded.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_blob_integrity(
                    env.download_bucket,
                    gcs_blob_name_from_path(release.download_country_path),
                    release.download_country_path,
                )
                self.assert_blob_integrity(
                    env.download_bucket,
                    gcs_blob_name_from_path(release.download_institution_path),
                    release.download_institution_path,
                )

                # Test that file transformed
                ti = env.run_task(telescope.transform.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assertTrue(os.path.exists(release.transform_country_path))
                self.assertTrue(os.path.exists(release.transform_institution_path))
                self.assert_file_integrity(
                    release.transform_country_path, self.country_report["transform_hash"], "gzip_crc"
                )
                self.assert_file_integrity(
                    release.transform_institution_path, self.institution_report["transform_hash"], "gzip_crc"
                )

                # Test that transformed file uploaded
                ti = env.run_task(telescope.upload_transformed.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_blob_integrity(
                    env.transform_bucket,
                    gcs_blob_name_from_path(release.transform_country_path),
                    release.transform_country_path,
                )
                self.assert_blob_integrity(
                    env.transform_bucket,
                    gcs_blob_name_from_path(release.transform_institution_path),
                    release.transform_institution_path,
                )

                # Test that data loaded into BigQuery
                ti = env.run_task(telescope.bq_load.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                country_table_id = bq_table_id(
                    telescope.cloud_workspace.project_id,
                    telescope.country_partner.bq_dataset_id,
                    telescope.country_partner.bq_table_name,
                )
                institution_table_id = bq_table_id(
                    telescope.cloud_workspace.project_id,
                    telescope.institution_partner.bq_dataset_id,
                    telescope.institution_partner.bq_table_name,
                )
                self.assert_table_integrity(country_table_id, self.country_report["table_rows"])
                self.assert_table_integrity(institution_table_id, self.institution_report["table_rows"])
                expected = load_and_parse_json(self.country_report["table"], date_fields=["release_date"])
                self.assert_table_content(country_table_id, expected, primary_key="ISBN")
                expected = load_and_parse_json(self.institution_report["table"], date_fields=["release_date"])
                self.assert_table_content(institution_table_id, expected, primary_key="ISBN")

                # Add_dataset_release_task
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 0)
                ti = env.run_task(telescope.add_new_dataset_releases.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 1)

                # Test that all telescope data deleted
                ti = env.run_task(telescope.cleanup.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_cleanup(release.workflow_folder)

    def test_get_release_date(self):
        """Test that the get_release_date returns the correct release date and raises an exception when dates are
        incorrect"""
        with CliRunner().isolated_filesystem():
            # Test reports in new format with header
            new_report_content = (
                'entity_name,book_title,item_doi,publisher,"Month, Year of monthdt"\n'
                "Entity,Book 1,10/1001,Publisher 1,{month_1}\n"  # Start date eg. September 2023
                "Entity,Book 2,10/1002,Publisher 1,{month_2}\n"
            )
            reports = [
                {"file": "new_success.tsv", "report_month_1": "January 2020", "report_month_2": "January 2020"},
                {"file": "new_fail.tsv", "report_month_1": "January 2020", "report_month_2": "February 2020"},
            ]

            for report in reports:
                with open(report["file"], "w") as f:
                    f.write(
                        new_report_content.format(month_1=report["report_month_1"], month_2=report["report_month_2"])
                    )

            # Test new report is successful
            api = JstorCollectionsAPI(Mock(), self.entity_id)
            start_date, end_date = api.get_release_date(reports[0]["file"])
            self.assertEqual(pendulum.parse("2020-01-01"), start_date)
            self.assertEqual(pendulum.parse("2020-01-31"), end_date)
            # Test new report fails
            with self.assertRaises(AirflowException):
                api.get_release_date(reports[1]["file"])


@patch("oaebu_workflows.jstor_telescope.jstor_telescope.build")
@patch("oaebu_workflows.jstor_telescope.jstor_telescope.Credentials")
def test_get_label_id(self, mock_account_credentials, mock_build):
    """Test getting label id both when label already exists and does not exist yet."""
    mock_account_credentials.from_json_keyfile_dict.return_value = ""

    http = HttpMockSequence(
        publisher_http_mock_sequence(
            self.country_report["url"], self.institution_report["url"], self.wrong_publisher_report["url"]
        )
    )
    mock_build.return_value = build("gmail", "v1", http=http)

    list_labels_no_match = {
        "labels": [
            {
                "id": "CHAT",
                "name": "CHAT",
                "messageListVisibility": "hide",
                "labelListVisibility": "labelHide",
                "type": "system",
            },
            {"id": "SENT", "name": "SENT", "type": "system"},
        ]
    }
    create_label = {
        "id": "created_label",
        "name": JstorTelescope.PROCESSED_LABEL_NAME,
        "messageListVisibility": "show",
        "labelListVisibility": "labelShow",
    }
    list_labels_match = {
        "labels": [
            {
                "id": "CHAT",
                "name": "CHAT",
                "messageListVisibility": "hide",
                "labelListVisibility": "labelHide",
                "type": "system",
            },
            {
                "id": "existing_label",
                "name": JstorTelescope.PROCESSED_LABEL_NAME,
                "messageListVisibility": "show",
                "labelListVisibility": "labelShow",
                "type": "user",
            },
        ]
    }
    http = HttpMockSequence(
        [
            ({"status": "200"}, json.dumps(list_labels_no_match)),
            ({"status": "200"}, json.dumps(create_label)),
            ({"status": "200"}, json.dumps(list_labels_match)),
        ]
    )
    service = build("gmail", "v1", http=http)
    api = JstorPublishersAPI(service, self.entity_id)

    # call function without match for label, so label is created
    label_id = api.get_label_id(service, JstorTelescope.PROCESSED_LABEL_NAME)
    self.assertEqual("created_label", label_id)

    # call function with match for label
    label_id = api.get_label_id(service, JstorTelescope.PROCESSED_LABEL_NAME)
    self.assertEqual("existing_label", label_id)


def publisher_http_mock_sequence(
    country_report_url: str, institution_report_url: str, wrong_publisher_report_url: str
) -> list:
    """Create a list with mocked http responses

    :param country_report_url: URL to country report
    :param institution_report_url: URL to institution report
    :param wrong_publisher_report_url: URL to report with a non-matching publisher id
    :return: List with http responses
    """
    list_labels = {
        "labels": [
            {
                "id": "CHAT",
                "name": "CHAT",
                "messageListVisibility": "hide",
                "labelListVisibility": "labelHide",
                "type": "system",
            },
            {
                "id": "Label_1",
                "name": JstorTelescope.PROCESSED_LABEL_NAME,
                "messageListVisibility": "show",
                "labelListVisibility": "labelShow",
                "type": "user",
            },
        ]
    }
    list_messages1 = {
        "messages": [
            {"id": "1788ec9e91f3de62", "threadId": "1788e9b0a848236a"},
        ],
        "resultSizeEstimate": 2,
        "nextPageToken": 1234,
    }
    list_messages2 = {
        "messages": [
            {"id": "1788ebe4ecbab055", "threadId": "1788e9b0a848236a"},
            {"id": "5621ayw3vjtag411", "threadId": "1788e9b0a848236a"},
        ],
        "resultSizeEstimate": 2,
    }
    get_message1 = {
        "id": "1788ec9e91f3de62",
        "threadId": "1788e9b0a848236a",
        "labelIds": ["CATEGORY_PERSONAL", "INBOX"],
        "snippet": "JSTOR JSTOR Usage Reports Report Complete Twitter Facebook Tumblr Dear OAEBU Service "
        "Account, Your usage report &quot;Book Usage by Country&quot; is now available to "
        "download. Download Completed Report",
        "payload": {
            "partId": "",
            "mimeType": "text/html",
            "filename": "",
            "headers": [{"name": "Delivered-To", "value": "accountname@gmail.com"}],
            "body": {
                "size": 12313,
                "data": base64.urlsafe_b64encode(
                    f'<a href="{country_report_url}">Download Completed Report</a>'.encode()
                ).decode(),
            },
        },
        "sizeEstimate": 17939,
        "historyId": "2302",
        "internalDate": "1617303299000",
    }
    get_message2 = {
        "id": "1788ebe4ecbab055",
        "threadId": "1788e9b0a848236a",
        "labelIds": ["CATEGORY_PERSONAL", "INBOX"],
        "snippet": "JSTOR JSTOR Usage Reports Report Complete Twitter Facebook Tumblr Dear OAEBU Service "
        "Account, Your usage report &quot;Book Usage by Country&quot; is now available to "
        "download. Download Completed Report",
        "payload": {
            "partId": "",
            "mimeType": "text/html",
            "filename": "",
            "headers": [{"name": "Delivered-To", "value": "accountname@gmail.com"}],
            "body": {
                "size": 12313,
                "data": base64.urlsafe_b64encode(
                    f'<a href="{institution_report_url}">Download Completed Report</a>'.encode()
                ).decode(),
            },
        },
        "sizeEstimate": 17939,
        "historyId": "2302",
        "internalDate": "1617303299000",
    }
    get_message3 = {
        "id": "5621ayw3vjtag411",
        "threadId": "1788e9b0a848236a",
        "labelIds": ["CATEGORY_PERSONAL", "INBOX"],
        "snippet": "JSTOR JSTOR Usage Reports Report Complete Twitter Facebook Tumblr Dear OAEBU Service "
        "Account, Your usage report &quot;Book Usage by Country&quot; is now available to "
        "download. Download Completed Report",
        "payload": {
            "partId": "",
            "mimeType": "text/html",
            "filename": "",
            "headers": [{"name": "Delivered-To", "value": "accountname@gmail.com"}],
            "body": {
                "size": 12313,
                "data": base64.urlsafe_b64encode(
                    f'<a href="{wrong_publisher_report_url}">Download Completed Report</a>'.encode()
                ).decode(),
            },
        },
        "sizeEstimate": 17939,
        "historyId": "2302",
        "internalDate": "1617303299000",
    }
    modify_message1 = {
        "id": "1788ec9e91f3de62",
        "threadId": "1788e9b0a848236a",
        "labelIds": ["Label_1", "CATEGORY_PERSONAL", "INBOX"],
    }
    modify_message2 = {
        "id": "1788ebe4ecbab055",
        "threadId": "1788e9b0a848236a",
        "labelIds": ["Label_1", "CATEGORY_PERSONAL", "INBOX"],
    }
    http_mock_sequence = [
        ({"status": "200"}, json.dumps(list_messages1)),
        ({"status": "200"}, json.dumps(list_messages2)),
        ({"status": "200"}, json.dumps(get_message1)),
        ({"status": "200"}, json.dumps(get_message2)),
        ({"status": "200"}, json.dumps(get_message3)),
        ({"status": "200"}, json.dumps(list_labels)),
        ({"status": "200"}, json.dumps(modify_message1)),
        ({"status": "200"}, json.dumps(modify_message2)),
    ]

    return http_mock_sequence


def collection_http_mock_sequence(country_json: str, institution_json: str) -> list:
    """Create a list with mocked http responses for the collection workflow of the telescope


    :param country_json: The path to the JSON file containing the data for the country message.
    :param institution_json: The path to the JSON file containing the data for the institution message.
    :returns: A list of tuples representing the HTTP mock responses
    """
    list_messages = {
        "messages": [
            {"id": "18af0b40b64fe408", "threadId": "18af0b3f613c3f95"},
            {"id": "18af0b3f613c3f95", "threadId": "18af0b3f613c3f95"},
        ],
        "resultSizeEstimate": 2,
    }

    get_message1 = {
        "id": "18af0b40b64fe408",
        "threadId": "18af0b3f613c3f95",
        "labelIds": ["CATEGORY_PERSONAL", "INBOX"],
        "snippet": "Monthly Usage of books by Country and Institution made available through the Big Ten Academic Alliance (BTAA) on JSTOR General Report Notes: Views and downloads of front matter, back matter, and table",
        "payload": {
            "partId": "",
            "mimeType": "multipart/mixed",
            "filename": "",
            "headers": [{"name": "Delivered-To", "value": "accountname@gmail.com"}],
            "body": {"size": 0},
            "parts": [
                {"partId": "0", "filename": "", "body": {"attachmentId": "0"}},
                {"partId": "1", "filename": "BTAA_Overall_Open_Usage.xlsx", "body": {"attachmentId": "1"}},
                {"partId": "2", "filename": "BTAA_Open_Country_Usage.csv", "body": {"attachmentId": "2"}},
                {"partId": "3", "filename": "BTAA_Open_Institution_Usage.csv", "body": {"attachmentId": "3"}},
            ],
        },
        "sizeEstimate": 355380,
        "historyId": "49808",
        "internalDate": "1696255445000",
    }
    get_message2 = {
        "id": "18af0b3f613c3f95",
        "threadId": "18af0b3f613c3f95",
        "labelIds": ["CATEGORY_PERSONAL", "INBOX"],
        "snippet": "Monthly Usage of books by Country and Institution made available through the Big Ten Academic Alliance (BTAA) on JSTOR General Report Notes: Views and downloads of front matter, back matter, and table",
        "payload": {
            "partId": "",
            "mimeType": "multipart/mixed",
            "filename": "",
            "headers": [{"name": "Delivered-To", "value": "accountname@gmail.com"}],
            "body": {"size": 0},
            # no "parts", this message should be ignored by the telescope
        },
        "sizeEstimate": 355380,
        "historyId": "49808",
        "internalDate": "1696255445000",
    }
    with open(country_json, "rb") as f:
        data_return1 = json.load(f)
        data_return1["data"] = base64.urlsafe_b64encode(data_return1["data"].encode()).decode()
    with open(institution_json, "rb") as f:
        data_return2 = json.load(f)
        data_return2["data"] = base64.urlsafe_b64encode(data_return2["data"].encode()).decode()

    list_labels = {
        "labels": [
            {
                "id": "CHAT",
                "name": "CHAT",
                "messageListVisibility": "hide",
                "labelListVisibility": "labelHide",
                "type": "system",
            },
            {
                "id": "Label_1",
                "name": JstorTelescope.PROCESSED_LABEL_NAME,
                "messageListVisibility": "show",
                "labelListVisibility": "labelShow",
                "type": "user",
            },
        ]
    }

    modify_message1 = {
        "id": "18af0b40b64fe408",
        "threadId": "18af0b3f613c3f95",
        "labelIds": ["Label_1", "CATEGORY_PERSONAL", "INBOX"],
    }

    http_mock_sequence = [
        ({"status": "200"}, json.dumps(list_messages)),
        ({"status": "200"}, json.dumps(get_message1)),
        ({"status": "200"}, json.dumps(get_message2)),
        ({"status": "200"}, json.dumps(data_return1)),
        ({"status": "200"}, json.dumps(data_return2)),
        ({"status": "200"}, json.dumps(list_labels)),
        ({"status": "200"}, json.dumps(modify_message1)),
    ]
    return http_mock_sequence



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, Keegan Smith

from __future__ import annotations

import logging
import os
import requests
import xmltodict
from xml.parsers.expat import ExpatError
from typing import Union

import pendulum
from airflow.exceptions import AirflowException
from google.cloud.bigquery import SourceFormat
from tenacity import (
    retry,
    stop_after_attempt,
    wait_chain,
    wait_fixed,
    retry_if_exception_type,
)

from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str
from oaebu_workflows.config import schema_folder
from oaebu_workflows.onix_utils import OnixTransformer
from observatory.api.client.model.dataset_release import DatasetRelease
from observatory.platform.api import make_observatory_api
from observatory.platform.airflow import AirflowConns
from observatory.platform.utils.url_utils import get_user_agent
from observatory.platform.observatory_config import CloudWorkspace
from observatory.platform.gcs import (
    gcs_upload_files,
    gcs_blob_uri,
    gcs_blob_name_from_path,
)
from observatory.platform.bigquery import bq_load_table, bq_sharded_table_id, bq_create_dataset
from observatory.platform.workflows.workflow import (
    SnapshotRelease,
    Workflow,
    make_snapshot_date,
    cleanup,
    set_task_state,
    check_workflow_inputs,
)


# Download job will wait 120 seconds between first 2 attempts, then 30 minutes for the following 3
DOWNLOAD_RETRY_CHAIN = wait_chain(*[wait_fixed(120) for _ in range(2)] + [wait_fixed(1800) for _ in range(3)])


class OapenMetadataRelease(SnapshotRelease):
    def __init__(self, dag_id: str, run_id: str, snapshot_date: pendulum.DateTime):
        """Construct a OapenMetadataRelease instance

        :param dag_id: The ID of the DAG
        :param run_id: The Airflow run ID
        :param snapshot_date: The date of the snapshot_date/release
        """
        super().__init__(dag_id=dag_id, run_id=run_id, snapshot_date=snapshot_date)
        self.download_path = os.path.join(self.download_folder, f"metadata_{snapshot_date.format('YYYYMMDD')}.xml")
        self.transform_path = os.path.join(self.transform_folder, "transformed.jsonl.gz")  # Final onix file

    @property
    def transform_files(self):
        files = os.listdir(self.transform_folder)
        return [os.path.join(self.transform_folder, f) for f in files]


class OapenMetadataTelescope(Workflow):
    """Oapen Metadata Telescope"""

    def __init__(
        self,
        dag_id: str,
        cloud_workspace: CloudWorkspace,
        metadata_uri: str,
        metadata_partner: Union[str, OaebuPartner] = "oapen_metadata",
        elevate_related_products: bool = False,
        bq_dataset_id: str = "onix",
        bq_table_name: str = "onix",
        bq_dataset_description: str = "OAPEN Metadata converted to ONIX",
        bq_table_description: str = None,
        api_dataset_id: str = "oapen",
        observatory_api_conn_id: str = AirflowConns.OBSERVATORY_API,
        catchup: bool = False,
        start_date: pendulum.DateTime = pendulum.datetime(2018, 5, 14),
        schedule: str = "0 12 * * Sun",  # Midday every sunday
    ):
        """Construct a OapenMetadataTelescope instance.
        :param dag_id: The ID of the DAG
        :param cloud_workspace: The CloudWorkspace object for this DAG
        :param metadata_uri: The URI of the metadata XML file
        :param metadata_partner: The metadata partner name
        :param elevate_related_products: Whether to pull out the related products to the product level.
        :param bq_dataset_id: The BigQuery dataset ID
        :param bq_table_name: The BigQuery table name
        :param bq_dataset_description: Description for the BigQuery dataset
        :param bq_table_description: Description for the biguery table
        :param api_dataset_id: The ID to store the dataset release in the API
        :param observatory_api_conn_id: Airflow connection ID for the overvatory API
        :param catchup: Whether to catchup the DAG or not
        :param start_date: The start date of the DAG
        :param schedule: The schedule interval of the DAG
        """
        super().__init__(
            dag_id,
            start_date,
            schedule,
            airflow_conns=[observatory_api_conn_id],
            catchup=catchup,
            tags=["oaebu"],
        )
        self.dag_id = dag_id
        self.cloud_workspace = cloud_workspace
        self.metadata_uri = metadata_uri
        self.metadata_partner = partner_from_str(metadata_partner, metadata_partner=True)
        self.elevate_related_products = elevate_related_products
        self.bq_dataset_id = bq_dataset_id
        self.bq_table_name = bq_table_name
        self.bq_dataset_description = bq_dataset_description
        self.bq_table_description = bq_table_description
        self.api_dataset_id = api_dataset_id
        self.observatory_api_conn_id = observatory_api_conn_id

        # Fixture file paths
        self.oapen_schema = os.path.join(
            schema_folder(workflow_module="oapen_metadata_telescope"), "oapen_metadata_filter.json"
        )

        check_workflow_inputs(self)

        self.add_setup_task(self.check_dependencies)
        self.add_task(self.download)
        self.add_task(self.upload_downloaded)
        self.add_task(self.transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.add_new_dataset_releases)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> OapenMetadataRelease:
        """Make release instances. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for the keyword arguments that can be passed
        :return: The Oapen metadata release instance"""
        snapshot_date = make_snapshot_date(**kwargs)
        release = OapenMetadataRelease(self.dag_id, kwargs["run_id"], snapshot_date)
        return release

    def download(self, release: OapenMetadataRelease, **kwargs) -> None:
        """Task to download the OapenMetadataRelease release.

        :param kwargs: the context passed from the PythonOperator.
        :param release: an OapenMetadataRelease instance.
        """
        logging.info(f"Downloading metadata XML from url: {self.metadata_uri}")
        download_metadata(self.metadata_uri, release.download_path)

    def upload_downloaded(self, release: OapenMetadataRelease, **kwargs) -> None:
        """Task to upload the downloaded OAPEN metadata"""
        success = gcs_upload_files(
            bucket_name=self.cloud_workspace.download_bucket,
            file_paths=[release.download_path],
        )
        set_task_state(success, kwargs["ti"].task_id, release=release)

    def transform(self, release: OapenMetadataRelease, **kwargs) -> None:
        """Transform the oapen metadata XML file into a valid ONIX file"""
        # Parse the downloaded metadata through the schema to extract relevant fields only
        transformer = OnixTransformer(
            input_path=release.download_path,
            output_dir=release.transform_folder,
            filter_products=True,
            error_removal=True,
            normalise_related_products=True,
            deduplicate_related_products=True,
            elevate_related_products=self.elevate_related_products,
            add_name_fields=True,
            collapse_subjects=True,
            filter_schema=self.oapen_schema,
            keep_intermediate=True,
        )
        out_file = transformer.transform()
        if release.transform_path != out_file:
            raise FileNotFoundError(f"Expected file {release.transform_path} not equal to transformed file: {out_file}")

    def upload_transformed(self, release: OapenMetadataRelease, **kwargs) -> None:
        """Task to upload the transformed OAPEN metadata"""
        success = gcs_upload_files(
            bucket_name=self.cloud_workspace.transform_bucket,
            file_paths=release.transform_files,
        )
        set_task_state(success, kwargs["ti"].task_id, release=release)

    def bq_load(self, release: OapenMetadataRelease, **kwargs) -> None:
        """Load the transformed ONIX file into bigquery"""
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.metadata_partner.bq_dataset_id,
            location=self.cloud_workspace.data_location,
            description=self.bq_dataset_description,
        )
        uri = gcs_blob_uri(
            self.cloud_workspace.transform_bucket,
            gcs_blob_name_from_path(release.transform_path),
        )
        table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.metadata_partner.bq_dataset_id,
            self.metadata_partner.bq_table_name,
            release.snapshot_date,
        )
        state = bq_load_table(
            uri=uri,
            table_id=table_id,
            schema_file_path=self.metadata_partner.schema_path,
            source_format=SourceFormat.NEWLINE_DELIMITED_JSON,
            table_description=self.bq_table_description,
        )
        set_task_state(state, kwargs["ti"].task_id, release=release)

    def add_new_dataset_releases(self, release: OapenMetadataRelease, **kwargs) -> None:
        """Adds release information to API."""
        api = make_observatory_api(observatory_api_conn_id=self.observatory_api_conn_id)
        dataset_release = DatasetRelease(
            dag_id=self.dag_id,
            dataset_id=self.api_dataset_id,
            dag_run_id=release.run_id,
            snapshot_date=release.snapshot_date,
            data_interval_start=kwargs["data_interval_start"],
            data_interval_end=kwargs["data_interval_end"],
        )
        api.post_dataset_release(dataset_release)

    def cleanup(self, release: OapenMetadataRelease, **kwargs) -> None:
        """Delete all files, folders and XComs associated with this release."""
        cleanup(
            dag_id=self.dag_id,
            execution_date=kwargs["execution_date"],
            workflow_folder=release.workflow_folder,
        )


@retry(
    stop=stop_after_attempt(5),
    wait=DOWNLOAD_RETRY_CHAIN,
    retry=retry_if_exception_type((ExpatError, ConnectionError, AirflowException)),
    reraise=True,
)
def download_metadata(uri: str, download_path: str) -> None:
    """Downloads the OAPEN metadata XML file
    OAPEN's downloader can give an incomplete file if the metadata is partially generated.
    In this scenario, we should wait until the metadata generator has finished.
    Otherwise, an attempt to parse the data will result in an XML ParseError.
    Another scenario is that OAPEN returns only a header in the XML. We want this to also raise an error.
    OAPEN metadata generation can take over an hour

    :param uri: the url to query for the metadata
    :param download_path: filepath to store te downloaded file
    :raises ConnectionError: raised if the response from the metadata server does not have code 200
    :raises AirflowException: raised if the response does not contain any Product fields
    """
    headers = {"User-Agent": f"{get_user_agent(package_name='oaebu_workflows')}"}
    response = requests.get(uri, headers=headers)
    if response.status_code != 200:
        raise ConnectionError(f"Expected status code 200 from url {uri}, instead got response: {response.text}")
    with open(download_path, "w") as f:
        f.write(response.content.decode("utf-8"))
    logging.info(f"Successfully downloadeded XML to {download_path}")

    # Attempt to parse the XML, will raise an ExpatError if it's invalid
    with open(download_path, "rb") as f:
        xmltodict.parse(f)
    logging.info("XML file is valid")

    # Check that more than just the header is returned
    if "<Product>" not in response.content.decode("utf-8"):
        raise AirflowException("No products found in metadata")



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Keegan Smith

from airflow.api.common.experimental.pool import create_pool, get_pool
from airflow.exceptions import PoolNotFound


class AirflowPool:
    def __init__(self, pool_name: str, pool_slots: int, pool_description: str):
        """Constructor an AirflowPool instance

        :param pool_name: The name of this pool
        :param pool_slots: The number of slots assigned to this pool
        :param pool_description: A description of this pool
        """
        self.pool_name = pool_name
        self.pool_slots = pool_slots
        self.pool_description = pool_description

    def get_pool(self):
        return get_pool(self.pool_name)

    def create_pool(self):
        return create_pool(self.pool_name, self.pool_slots, self.pool_description)

    def create_or_get_pool(self):
        try:
            return self.get_pool()
        except PoolNotFound:
            return self.create_pool()


class CrossrefEventsPool(AirflowPool):
    def __init__(self, pool_slots: int = 15):
        """Constructor CrossrefEventsPool instance

        :param pool_slots: The number of slots assigned to this pool
        """
        super().__init__(
            pool_name="Crossref Events Pool", pool_slots=pool_slots, pool_description="Crossref Events API Pool"
        )
